#!/usr/bin/python3

# ibus-typing-booster - A completion input method for IBus
#
# Copyright (c) 2019 Mike FABIAN <mfabian@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

'''
This file implements test cases for emoji candidates
'''

import sys
import os
import importlib.util
import logging
import unittest

LOGGER = logging.getLogger('ibus-typing-booster')

# pylint: disable=wrong-import-position
sys.path.insert(0, "../engine")
import itb_emoji # pylint: disable=import-error
sys.path.pop(0)

import testutils # pylint: disable=import-error
# pylint: enable=wrong-import-position

# Set the domain name to something invalid to avoid using
# the translations for the doctest tests. Translations may
# make the tests fail just because some translations are
# added, changed, or missing.
itb_emoji.DOMAINNAME = ''

IS_ENCHANT_AVAILABLE = importlib.util.find_spec('enchant') is not None

# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=invalid-name
# pylint: disable=line-too-long

@unittest.skipIf(
    os.path.abspath(os.path.join(os.path.dirname(__file__), '../data/annotations'))
    != os.path.abspath(os.path.dirname(itb_emoji.find_cldr_annotation_path('en'))),
    f'Using external emoji annotations: '
    f'{itb_emoji.find_cldr_annotation_path("en")} '
    f'Testing with older emoji annotations instead '
    f'of those included in the ibus-typing-booster source is likely '
    f'to create meaningless test failures.')
class EmojiCandidatesTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.maxDiff = None
        LOGGER.info("itb_emoji.find_cldr_annotation_path('en')->%s",
                    itb_emoji.find_cldr_annotation_path('en'))

    def tearDown(self) -> None:
        pass

    def test_dummy(self) -> None:
        self.assertEqual(True, True)

    def test_candidates_empty_query(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['fr_FR'])
        self.assertEqual(mq.candidates(''), [])

    def test_candidates_empty_language_list(self) -> None:
        '''
        Emoji matching with an empty language list should fall back to English.
        '''
        mq = itb_emoji.EmojiMatcher(
            languages = [])
        self.assertEqual(
            mq.candidates('orangutan', match_limit=1)[0].phrase,
            '🦧')
        mq = itb_emoji.EmojiMatcher(
            languages = ['en'])
        self.assertEqual(
            mq.candidates('orangutan', match_limit=1)[0].phrase,
            '🦧')

    def test_candidates_similar_emoji(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US', 'it_IT', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'])
        matches = mq.candidates('😺', match_limit=3)
        self.assertEqual(matches[0].phrase, '😺')
        self.assertEqual(matches[0].comment, 'smiling cat face with open mouth [😺, So, people, cat, face, mouth, open, smile, uc6, animal, grinning, smiling]')
        self.assertEqual(matches[0].user_freq, 12.0)
        self.assertEqual(matches[1].phrase, '😸')
        self.assertEqual(matches[1].comment, 'grinning cat face with smiling eyes [So, people, cat, face, smile, uc6, animal, grinning, smiling]')
        self.assertEqual(matches[1].user_freq, 9.0)
        self.assertEqual(matches[2].phrase, '😅')
        self.assertEqual(matches[2].comment, 'smiling face with open mouth and cold sweat [So, people, face, open, smile, uc6, grinning, mouth, smiling]')
        self.assertEqual(matches[2].user_freq, 9.0)

    def test_candidates_japanese_full_width_low_line(self) -> None:
        # ＿ U+FF3F FULLWIDTH LOW LINE should not disturb the match
        mq = itb_emoji.EmojiMatcher(languages = ['ja_JP'])
        first_match = mq.candidates('ネコ')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'ネコ')
        first_match = mq.candidates('ネコ＿')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'ネコ')

    def test_candidates_multilingual_classic(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US', 'it_IT', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'],
            match_algorithm='classic')
        first_match = mq.candidates('ant')[0]
        self.assertEqual(first_match.phrase, '🐜')
        self.assertEqual(first_match.comment, 'ant')
        first_match = mq.candidates('ameise')[0]
        self.assertEqual(first_match.phrase, '🐜')
        self.assertEqual(first_match.comment, 'Ameise')
        first_match = mq.candidates('Ameise')[0]
        self.assertEqual(first_match.phrase, '🐜')
        self.assertEqual(first_match.comment, 'Ameise')
        first_match = mq.candidates('formica')[0]
        self.assertEqual(first_match.phrase, '🐜')
        self.assertEqual(first_match.comment, 'formica')
        first_match = mq.candidates('hormiga')[0]
        self.assertEqual(first_match.phrase, '🐜')
        self.assertEqual(first_match.comment, 'hormiga')
        first_match = mq.candidates('cacca')[0]
        self.assertEqual(first_match.phrase, '💩')
        self.assertEqual(first_match.comment, 'cacca')
        first_match = mq.candidates('orso')[0]
        self.assertEqual(first_match.phrase, '🐻')
        self.assertEqual(first_match.comment, 'orso')
        first_match = mq.candidates('lupo')[0]
        self.assertEqual(first_match.phrase, '🐺')
        self.assertEqual(first_match.comment, 'lupo')
        first_match = mq.candidates('gatto')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'gatto')
        first_match = mq.candidates('gatto sorride')[0]
        self.assertEqual(first_match.phrase, '😺')
        self.assertEqual(first_match.comment, 'gatto che sorride')
        first_match = mq.candidates('halo')[0]
        self.assertEqual(first_match.phrase, '😇')
        self.assertEqual(first_match.comment, 'cara sonriendo con aureola [halo]')
        first_match = mq.candidates('factory')[0]
        self.assertEqual(first_match.phrase, '🏭')
        self.assertEqual(first_match.comment, 'factory')

    @unittest.skipUnless(
        itb_emoji.IMPORT_RAPIDFUZZ_SUCCESSFUL,
        'Skipping because this test requires rapidfuzz to work.')
    def test_candidates_multilingual_rapidfuzz(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US', 'it_IT', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'],
            match_algorithm = 'rapidfuzz')
        first_match = mq.candidates('ant')[0]
        self.assertEqual(first_match.phrase, '🐜')
        self.assertEqual(first_match.comment, 'ant')
        first_match = mq.candidates('ameise')[0]
        self.assertEqual(first_match.phrase, '🐜')
        self.assertEqual(first_match.comment, 'Ameise')
        first_match = mq.candidates('Ameise')[0]
        self.assertEqual(first_match.phrase, '🐜', 'Ameise')
        first_match = mq.candidates('formica')[0]
        self.assertEqual(first_match.phrase, '🐜')
        self.assertEqual(first_match.comment, 'formica [formichina]')
        first_match = mq.candidates('hormiga')[0]
        self.assertEqual(first_match.phrase, '🐜')
        self.assertEqual(first_match.comment, 'hormiga')
        first_match, second_match = mq.candidates('cacca')[:2]
        self.assertEqual(first_match.phrase, '💩')
        self.assertEqual(first_match.comment, 'cacca')
        self.assertEqual(second_match.phrase, '\U000cacca')
        self.assertEqual(second_match.comment, 'U+CACCA')
        first_match = mq.candidates('orso')[0]
        self.assertEqual(first_match.phrase, '🐻')
        self.assertEqual(first_match.comment, 'orso')
        first_match = mq.candidates('lupo')[0]
        self.assertEqual(first_match.phrase, '🐺')
        self.assertEqual(first_match.comment, 'lupo [muso di lupo]')
        first_match = mq.candidates('gatto')[0]
        self.assertEqual(first_match.phrase, '😺')
        self.assertEqual(first_match.comment, 'gatto che sorride')
        first_match = mq.candidates('gatto sorride')[0]
        self.assertEqual(first_match.phrase, '😺')
        self.assertEqual(first_match.comment, 'gatto che sorride [sorridente]')
        first_match = mq.candidates('halo')[0]
        self.assertEqual(first_match.phrase, '😇')
        self.assertEqual(first_match.comment, 'smiling face with halo')
        first_match = mq.candidates('factory')[0]
        self.assertEqual(first_match.phrase, '🧑\u200d🏭')
        self.assertEqual(first_match.comment, 'factory worker')

    def test_candidates_white_space_and_underscores_classic(self) -> None:
        # Any white space and '_' can be used to separate keywords in the
        # query string:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US', 'it_IT', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'],
            match_algorithm='classic')
        first_match = mq.candidates('gatto_	 sorride')[0]
        self.assertEqual(first_match.phrase, '😺')
        self.assertEqual(first_match.comment, 'gatto che sorride')
        first_match = mq.candidates('nerd glasses')[0]
        self.assertEqual(first_match.phrase, '🤓')
        self.assertEqual(first_match.comment, 'nerd face [glasses]')
        first_match = mq.candidates('smiling face with sunglasses')[0]
        self.assertEqual(first_match.phrase, '😎')
        self.assertEqual(first_match.comment, 'smiling face with sunglasses')

    @unittest.skipUnless(
        itb_emoji.IMPORT_RAPIDFUZZ_SUCCESSFUL,
        'Skipping because this test requires rapidfuzz to work.')
    def test_candidates_white_space_and_underscores_rapidfuzz(self) -> None:
        # Any white space and '_' can be used to separate keywords in the
        # query string:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US', 'it_IT', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'],
            match_algorithm='rapidfuzz')
        first_match = mq.candidates('gatto_	 sorride')[0]
        self.assertEqual(first_match.phrase, '😺')
        self.assertEqual(first_match.comment, 'gatto che sorride [sorridente]')
        first_match = mq.candidates('nerd glasses')[0]
        self.assertEqual(first_match.phrase, '🤓')
        self.assertEqual(first_match.comment, 'nerd face [glasses]')
        first_match = mq.candidates('smiling face with sunglasses')[0]
        self.assertEqual(first_match.phrase, '😎')
        self.assertEqual(first_match.comment, 'smiling face with sunglasses')

    def test_candidates_skin_tones_classic(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['ja_JP'], match_algorithm='classic')
        first_match = mq.candidates('man tone5')[0]
        self.assertEqual(first_match.phrase, '👨🏿')
        self.assertEqual(first_match.comment, 'man: dark skin tone “man tone5”')
        first_match = mq.candidates('skin tone')[0]
        self.assertEqual(first_match.phrase, '👋🏻')
        self.assertEqual(first_match.comment, 'waving hand: light skin tone')
        first_match = mq.candidates('tone1')[0]
        self.assertEqual(first_match.phrase, '🏻')
        self.assertEqual(first_match.comment, 'emoji modifier fitzpatrick type-1-2 “tone1”')
        first_match = mq.candidates('tone5')[0]
        self.assertEqual(first_match.phrase, '🏿')
        self.assertEqual(first_match.comment, 'emoji modifier fitzpatrick type-6 “tone5”')

    @unittest.skipUnless(
        itb_emoji.IMPORT_RAPIDFUZZ_SUCCESSFUL,
        'Skipping because this test requires rapidfuzz to work.')
    def test_candidates_skin_tones_rapidfuzz(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['ja_JP'], match_algorithm='rapidfuzz')
        first_match = mq.candidates('man tone5')[0]
        self.assertEqual(first_match.phrase, '👨🏿')
        self.assertEqual(first_match.comment, 'man: dark skin tone “man tone5”')
        first_match = mq.candidates('skin tone')[0]
        self.assertEqual(first_match.phrase, '👋🏻')
        self.assertEqual(first_match.comment, 'waving hand: light skin tone')
        first_match = mq.candidates('tone1')[0]
        self.assertEqual(first_match.phrase, '👋🏻')
        self.assertEqual(first_match.comment, 'waving hand: light skin tone “wave tone1”')
        first_match = mq.candidates('tone5')[0]
        self.assertEqual(first_match.phrase, '👋🏿')
        self.assertEqual(first_match.comment, 'waving hand: dark skin tone “wave tone5”')

    def test_candidates_some_letters_classic(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US', 'it_IT', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'],
            match_algorithm='classic')
        first_match = mq.candidates('latin capital letter a')[0]
        self.assertEqual(first_match.phrase, '🅰\ufe0f')
        self.assertEqual(first_match.comment, 'negative squared latin capital letter a')
        first_match = mq.candidates('squared a')[0]
        self.assertEqual(first_match.phrase, '🅰\ufe0f')
        self.assertEqual(first_match.comment, 'negative squared latin capital letter a')
        first_match = mq.candidates('squared capital a')[0]
        self.assertEqual(first_match.phrase, '🅰\ufe0f')
        self.assertEqual(first_match.comment, 'negative squared latin capital letter a')
        first_match = mq.candidates('latin small letter c')[0]
        self.assertEqual(first_match.phrase, '🔡')
        self.assertEqual(first_match.comment, 'input symbol for latin small letters')

    @unittest.skipUnless(
        itb_emoji.IMPORT_RAPIDFUZZ_SUCCESSFUL,
        'Skipping because this test requires rapidfuzz to work.')
    def test_candidates_some_letters_rapidfuzz(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US', 'it_IT', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'],
            match_algorithm='rapidfuzz')
        first_match = mq.candidates('latin capital letter a')[0]
        self.assertEqual(first_match.phrase, '🅰\ufe0f')
        self.assertEqual(first_match.comment, 'negative squared latin capital letter a')
        first_match = mq.candidates('squared a')[0]
        self.assertEqual(first_match.phrase, '🅰\ufe0f')
        self.assertEqual(first_match.comment, 'negative squared latin capital letter a')
        first_match = mq.candidates('squared capital a')[0]
        self.assertEqual(first_match.phrase, '🅰\ufe0f')
        self.assertEqual(first_match.comment, 'negative squared latin capital letter a')
        first_match = mq.candidates('latin small letter c')[0]
        self.assertEqual(first_match.phrase, '🔡')
        self.assertEqual(first_match.comment, 'input symbol for latin small letters')

    def test_candidates_flags_classic(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US'], match_algorithm='classic')
        first_match = mq.candidates('us')[0]
        self.assertEqual(first_match.phrase, '🇺🇸')
        self.assertEqual(first_match.comment, 'flag: united states “us”')
        first_match = mq.candidates('flag us')[0]
        self.assertEqual(first_match.phrase, '🇺🇸')
        self.assertEqual(first_match.comment, 'flag: united states “flag us”')
        first_match = mq.candidates('united nations')[0]
        self.assertEqual(first_match.phrase, '🇺🇳')
        self.assertEqual(first_match.comment, 'flag: united nations')
        first_match = mq.candidates('united')[0]
        self.assertEqual(first_match.phrase, '🇺🇳')
        self.assertEqual(first_match.comment, 'flag: united nations')
        first_match = mq.candidates('outlying islands')[0]
        self.assertEqual(first_match.phrase, '🇺🇲')
        self.assertEqual(first_match.comment, 'flag: u.s. outlying islands')
        first_match = mq.candidates('flag united arab')[0]
        self.assertEqual(first_match.phrase, '🇦🇪')
        self.assertEqual(first_match.comment, 'flag: united arab emirates')
        first_match = mq.candidates('mm')[0]
        self.assertEqual(first_match.phrase, '🇲🇲')
        self.assertEqual(first_match.comment, 'flag: myanmar (burma) “mm”')
        first_match = mq.candidates('flag mm')[0]
        self.assertEqual(first_match.phrase, '🇲🇲')
        self.assertEqual(first_match.comment, 'flag: myanmar (burma) “flag mm”')
        first_match = mq.candidates('myanmar burma')[0]
        self.assertEqual(first_match.phrase, '🇲🇲')
        self.assertEqual(first_match.comment, 'flag: myanmar (burma) “flag: myanmar burma”')
        first_match = mq.candidates('sj')[0]
        self.assertEqual(first_match.phrase, '🇸🇯')
        self.assertEqual(first_match.comment, 'flag: svalbard & jan mayen “sj”')
        first_match = mq.candidates('flag sj')[0]
        self.assertEqual(first_match.phrase, '🇸🇯')
        self.assertEqual(first_match.comment, 'flag: svalbard & jan mayen “flag sj”')
        first_match = mq.candidates('svalbard')[0]
        self.assertEqual(first_match.phrase, '🇸🇯')
        self.assertEqual(first_match.comment, 'flag: svalbard & jan mayen')
        first_match = mq.candidates('jan mayen')[0]
        self.assertEqual(first_match.phrase, '🇸🇯')
        self.assertEqual(first_match.comment, 'flag: svalbard & jan mayen')
        first_match = mq.candidates('mayen')[0]
        self.assertEqual(first_match.phrase, '🇸🇯')
        self.assertEqual(first_match.comment, 'flag: svalbard & jan mayen')

    @unittest.skipUnless(
        itb_emoji.IMPORT_RAPIDFUZZ_SUCCESSFUL,
        'Skipping because this test requires rapidfuzz to work.')
    def test_candidates_flags_rapidfuzz(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US'], nameslist=False, match_algorithm='rapidfuzz')
        first_match = mq.candidates('us')[0]
        self.assertEqual(first_match.phrase, '🇺🇲')
        self.assertEqual(first_match.comment, 'flag: us outlying islands')
        first_match = mq.candidates('flag us')[0]
        self.assertEqual(first_match.phrase, '🇺🇸')
        self.assertEqual(first_match.comment, 'flag: united states “flag us” {flags}')
        first_match = mq.candidates('united nations')[0]
        self.assertEqual(first_match.phrase, '🇺🇳')
        self.assertEqual(first_match.comment, 'flag: united nations')
        first_match = mq.candidates('united')[0]
        self.assertEqual(first_match.phrase, '🇦🇪')
        self.assertEqual(first_match.comment, 'flag: united arab emirates')
        first_match = mq.candidates('outlying islands')[0]
        self.assertEqual(first_match.phrase, '🇺🇲')
        self.assertEqual(first_match.comment, 'flag: u.s. outlying islands')
        first_match = mq.candidates('flag united arab emirates')[0]
        self.assertEqual(first_match.phrase, '🇦🇪')
        self.assertEqual(first_match.comment, 'flag: united arab emirates')
        first_match = mq.candidates('mm myanmar')[0]
        self.assertEqual(first_match.phrase, '🇲🇲')
        self.assertEqual(first_match.comment, 'flag: myanmar (burma)')
        first_match = mq.candidates('flag mm')[0]
        self.assertEqual(first_match.phrase, '🇲🇲')
        self.assertEqual(first_match.comment, 'flag: myanmar (burma) “flag mm” {flags}')
        first_match = mq.candidates('myanmar')[0]
        self.assertEqual(first_match.phrase, '🇲🇲')
        self.assertEqual(first_match.comment, 'flag: myanmar (burma)')
        first_match = mq.candidates('sj')[0]
        self.assertEqual(first_match.phrase, '🇸🇯')
        self.assertEqual(first_match.comment, 'flag: svalbard & jan mayen “flag sj”')
        first_match = mq.candidates('flag sj')[0]
        self.assertEqual(first_match.phrase, '🇸🇯')
        self.assertEqual(first_match.comment, 'flag: svalbard & jan mayen “flag sj” {flags}')
        first_match = mq.candidates('svalbard')[0]
        self.assertEqual(first_match.phrase, '🇸🇯')
        self.assertEqual(first_match.comment, 'flag: svalbard & jan mayen')
        first_match = mq.candidates('jan mayen')[0]
        self.assertEqual(first_match.phrase, '🇸🇯')
        self.assertEqual(first_match.comment, 'flag: svalbard & jan mayen')
        first_match = mq.candidates('mayen')[0]
        self.assertEqual(first_match.phrase, '🇸🇯')
        self.assertEqual(first_match.comment, 'flag: svalbard & jan mayen')

    @unittest.skipUnless(
        IS_ENCHANT_AVAILABLE,
        "Skipping because this test requires python3-enchant to work.")
    @unittest.skipUnless(
        testutils.enchant_working_as_expected(),
        'Skipping because of an unexpected change in the enchant behaviour.')
    def test_candidates_persons_classic(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US'], match_algorithm='classic')
        first_match = mq.candidates('family')[0]
        self.assertEqual(first_match.phrase, '👪')
        self.assertEqual(first_match.comment, 'family')
        first_match = mq.candidates('man')[0]
        self.assertEqual(first_match.phrase, '👨')
        self.assertEqual(first_match.comment, 'man')
        first_match = mq.candidates('woman')[0]
        self.assertEqual(first_match.phrase, '👩')
        self.assertEqual(first_match.comment, 'woman')
        first_match = mq.candidates('girl')[0]
        self.assertEqual(first_match.phrase, '👧')
        self.assertEqual(first_match.comment, 'girl')
        first_match = mq.candidates('boy')[0]
        self.assertEqual(first_match.phrase, '👦')
        self.assertEqual(first_match.comment, 'boy')
        first_match = mq.candidates('family man')[0]
        self.assertEqual(first_match.phrase, '👨\u200d👩\u200d👦')
        self.assertEqual(first_match.comment, 'family: man, woman, boy')
        first_match = mq.candidates('man woman girl boy')[0]
        self.assertEqual(first_match.phrase, '👨‍👩‍👧‍👦')
        self.assertEqual(first_match.comment, 'family: man, woman, girl, boy')
        first_match = mq.candidates('people holding hands')[0]
        self.assertEqual(first_match.phrase, '🧑‍🤝‍🧑')
        self.assertEqual(first_match.comment, 'people holding hands')

    @unittest.skipUnless(
        IS_ENCHANT_AVAILABLE,
        "Skipping because this test requires python3-enchant to work.")
    @unittest.skipUnless(
        testutils.enchant_working_as_expected(),
        'Skipping because of an unexpected change in the enchant behaviour.')
    @unittest.skipUnless(
        itb_emoji.IMPORT_RAPIDFUZZ_SUCCESSFUL,
        'Skipping because this test requires rapidfuzz to work.')
    def test_candidates_persons_rapidfuzz(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US'], match_algorithm='rapidfuzz')
        first_match = mq.candidates('family people')[0]
        self.assertEqual(first_match.phrase, '👪')
        self.assertEqual(first_match.comment, 'family {people}')
        first_match = mq.candidates('man people')[0]
        self.assertEqual(first_match.phrase, '👨')
        self.assertEqual(first_match.comment, 'man {people}')
        first_match = mq.candidates('woman people')[0]
        self.assertEqual(first_match.phrase, '👩')
        self.assertEqual(first_match.comment, 'woman {people}')
        first_match = mq.candidates('girl people')[0]
        self.assertEqual(first_match.phrase, '👧')
        self.assertEqual(first_match.comment, 'girl {people}')
        first_match = mq.candidates('boy people')[0]
        self.assertEqual(first_match.phrase, '👦')
        self.assertEqual(first_match.comment, 'boy {people}')
        first_match = mq.candidates('family man')[0]
        self.assertEqual(first_match.phrase, '👨\u200d👩\u200d👦')
        self.assertEqual(first_match.comment, 'family: man, woman, boy')
        first_match = mq.candidates('man woman girl boy')[0]
        self.assertEqual(first_match.phrase, '👨‍👩‍👧‍👦')
        self.assertEqual(first_match.comment, 'family: man, woman, girl, boy')
        first_match = mq.candidates('people holding hands')[0]
        self.assertEqual(first_match.phrase, '🧑‍🤝‍🧑')
        self.assertEqual(first_match.comment, 'people holding hands')

    def test_candidates_birthday_cake_classic(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US', 'it_IT', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'],
            match_algorithm = 'classic')
        first_match = mq.candidates('birthday')[0]
        self.assertEqual(first_match.phrase, '🎂')
        self.assertEqual(first_match.comment, 'birthday cake')
        first_match = mq.candidates('birth bottle')[0]
        self.assertEqual(first_match.phrase, '🍼')
        self.assertEqual(first_match.comment, 'baby bottle [birth]')

    @unittest.skipUnless(
        itb_emoji.IMPORT_RAPIDFUZZ_SUCCESSFUL,
        'Skipping because this test requires rapidfuzz to work.')
    def test_candidates_birthday_cake_rapidfuzz(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US', 'it_IT', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'],
            match_algorithm = 'rapidfuzz')
        first_match = mq.candidates('birthday')[0]
        self.assertEqual(first_match.phrase, '🎂')
        self.assertEqual(first_match.comment, 'birthday cake [bday]')
        first_match = mq.candidates('birth bottle')[0]
        self.assertEqual(first_match.phrase, '🍼')
        self.assertEqual(first_match.comment, 'baby bottle [birth]')

    @unittest.skipUnless(
        IS_ENCHANT_AVAILABLE,
        "Skipping because this test requires python3-enchant to work.")
    @unittest.skipUnless(
        testutils.enchant_working_as_expected(),
        'Skipping because of an unexpected change in the enchant behaviour.')
    def test_candidates_symbols_classic(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US', 'it_IT', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'],
            match_algorithm='classic')
        first_match = mq.candidates('recycling symbol')[0]
        self.assertEqual(first_match.phrase, '♻️')
        self.assertEqual(first_match.comment,
                         'black universal recycling symbol {Symbol}')
        first_match = mq.candidates('atom symbol')[0]
        self.assertEqual(first_match.phrase, '⚛\ufe0f')
        self.assertEqual(first_match.comment, 'atom symbol {Symbol}')
        first_match = mq.candidates('peace symbol')[0]
        self.assertEqual(first_match.phrase, '☮\ufe0f')
        # .startswith() because it may be 'peace symbol {Symbol}' or
        # just 'peace symbol'
        self.assertTrue(first_match.comment.startswith('peace symbol'))
        first_match = mq.candidates('peace symbol')[0]
        self.assertEqual(first_match.phrase, '☮\ufe0f')
        # .startswith() because it may be 'peace symbol {Symbol}' or
        # just 'peace symbol'
        self.assertTrue(first_match.comment.startswith('peace symbol'))

    @unittest.skipUnless(
        IS_ENCHANT_AVAILABLE,
        "Skipping because this test requires python3-enchant to work.")
    @unittest.skipUnless(
        testutils.enchant_working_as_expected(),
        'Skipping because of an unexpected change in the enchant behaviour.')
    @unittest.skipUnless(
        itb_emoji.IMPORT_RAPIDFUZZ_SUCCESSFUL,
        'Skipping because this test requires rapidfuzz to work.')
    def test_candidates_symbols_rapidfuzz(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US', 'it_IT', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'],
            nameslist=False,
            match_algorithm='rapidfuzz')
        first_match = mq.candidates('recycling symbol')[0]
        self.assertEqual(first_match.phrase, '♻️')
        self.assertEqual(first_match.comment,
                         'black universal recycling symbol {Symbol}')
        first_match = mq.candidates('atom symbol')[0]
        self.assertEqual(first_match.phrase, '⚛\ufe0f')
        self.assertEqual(first_match.comment, 'atom symbol {Symbol} {symbols}')
        first_match = mq.candidates('peace symbol')[0]
        self.assertEqual(first_match.phrase, '☮\ufe0f')
        # .startswith() because it may be 'peace symbol {Symbol}' or
        # just 'peace symbol'
        self.assertTrue(first_match.comment.startswith('peace symbol'))
        first_match = mq.candidates('peace symbol')[0]
        self.assertEqual(first_match.phrase, '☮\ufe0f')
        # .startswith() because it may be 'peace symbol {Symbol}' or
        # just 'peace symbol'
        self.assertTrue(first_match.comment.startswith('peace symbol'))

    def test_candidates_animals_classic(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US', 'it_IT', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'],
            match_algorithm='classic')
        first_match = mq.candidates('animal dog')[0]
        self.assertEqual(first_match.phrase, '🐶')
        self.assertEqual(first_match.comment, 'dog face [animal]')
        first_match = mq.candidates('dromedary animal')[0]
        self.assertEqual(first_match.phrase, '🐪')
        self.assertEqual(first_match.comment, 'dromedary camel [animal]')
        first_match = mq.candidates('camel bac')[0]
        self.assertEqual(first_match.phrase, '🐫')
        self.assertEqual(first_match.comment, 'bactrian camel')
        first_match = mq.candidates('nature snail')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'snail {nature}')

    @unittest.skipUnless(
        itb_emoji.IMPORT_RAPIDFUZZ_SUCCESSFUL,
        'Skipping because this test requires rapidfuzz to work.')
    def test_candidates_animals_rapidfuzz(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US', 'it_IT', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'],
            match_algorithm='rapidfuzz')
        first_match = mq.candidates('animal dog')[0]
        self.assertEqual(first_match.phrase, '🐶')
        self.assertEqual(first_match.comment, 'dog face [animal]')
        first_match = mq.candidates('dromedary animal')[0]
        self.assertEqual(first_match.phrase, '🐪')
        self.assertEqual(first_match.comment, 'dromedary camel [animal]')
        first_match = mq.candidates('camel bac')[0]
        self.assertEqual(first_match.phrase, '🐫')
        self.assertEqual(first_match.comment, 'bactrian camel')
        first_match = mq.candidates('nature snail')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'snail {nature}')

    def test_candidates_travel_classic(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US', 'it_IT', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'],
            nameslist=False,
            match_algorithm='classic')
        first_match = mq.candidates('camera')[0]
        self.assertEqual(first_match.phrase, '📷')
        self.assertEqual(first_match.comment, 'camera')
        first_match = mq.candidates('travel locomotive')[0]
        self.assertEqual(first_match.phrase, '🚂')
        self.assertEqual(first_match.comment, 'steam locomotive {travel}')
        first_match = mq.candidates('ferry')[0]
        self.assertEqual(first_match.phrase, '⛴\ufe0f')
        self.assertEqual(first_match.comment, 'ferry')
        first_match = mq.candidates('ferry travel')[0]
        self.assertEqual(first_match.phrase, '⛴\ufe0f')
        self.assertEqual(first_match.comment, 'ferry {travel}')
        first_match = mq.candidates('ferry travel boat')[0]
        self.assertEqual(first_match.phrase, '⛴\ufe0f')
        self.assertEqual(first_match.comment, 'ferry {travel}')
        first_match = mq.candidates('boat')[0]
        self.assertEqual(first_match.phrase, '🚣')
        self.assertEqual(first_match.comment, 'rowboat')
        first_match = mq.candidates('anchor')[0]
        self.assertEqual(first_match.phrase, '⚓')
        self.assertEqual(first_match.comment, 'anchor')
        first_match = mq.candidates('anchor ship')[0]
        self.assertEqual(first_match.phrase, '⚓')
        self.assertEqual(first_match.comment, 'anchor')

    @unittest.skipUnless(
        itb_emoji.IMPORT_RAPIDFUZZ_SUCCESSFUL,
        'Skipping because this test requires rapidfuzz to work.')
    def test_candidates_travel_rapidfuzz(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US', 'it_IT', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'],
            match_algorithm='rapidfuzz')
        first_match = mq.candidates('camera')[0]
        self.assertEqual(first_match.phrase, '🎥')
        self.assertEqual(first_match.comment, 'movie camera')
        first_match = mq.candidates('travel locomotive')[0]
        self.assertEqual(first_match.phrase, '🚂')
        self.assertEqual(first_match.comment, 'steam locomotive {travel}')
        first_match = mq.candidates('ferry')[0]
        self.assertEqual(first_match.phrase, '⛴\ufe0f')
        self.assertEqual(first_match.comment, 'ferry')
        first_match = mq.candidates('ferry travel')[0]
        self.assertEqual(first_match.phrase, '⛴\ufe0f')
        self.assertEqual(first_match.comment, 'ferry {travel}')
        first_match = mq.candidates('ferry travel boat')[0]
        self.assertEqual(first_match.phrase, '⛴\ufe0f')
        self.assertEqual(first_match.comment, 'ferry {travel}')
        first_match = mq.candidates('boat')[0]
        self.assertEqual(first_match.phrase, '🚣')
        self.assertEqual(first_match.comment, 'rowboat')
        first_match = mq.candidates('anchor')[0]
        self.assertEqual(first_match.phrase, '⚓')
        self.assertEqual(first_match.comment, 'anchor')
        first_match = mq.candidates('anchor ship')[0]
        self.assertEqual(first_match.phrase, '⚓')
        self.assertEqual(first_match.comment, 'anchor')

    @unittest.skipUnless(
        IS_ENCHANT_AVAILABLE,
        "Skipping because this test requires python3-enchant to work.")
    @unittest.skipUnless(
        testutils.enchant_working_as_expected(),
        'Skipping because of an unexpected change in the enchant behaviour.')
    def test_candidates_spellchecking(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US'])
        first_match = mq.candidates('buterfly', spellcheck=True)[0]
        self.assertEqual(first_match.phrase, '\U0001f98b')
        self.assertEqual(first_match.comment, 'butterfly')
        first_match = mq.candidates('badminton', spellcheck=True)[0]
        self.assertEqual(first_match.phrase, '🏸')
        self.assertEqual(first_match.comment, 'badminton racquet and shuttlecock')
        first_match = mq.candidates('badmynton', spellcheck=True)[0]
        self.assertEqual(first_match.phrase, '🏸')
        self.assertEqual(first_match.comment, 'badminton racquet and shuttlecock')
        first_match = mq.candidates('padminton', spellcheck=True)[0]
        self.assertEqual(first_match.phrase, '🏸')
        self.assertEqual(first_match.comment, 'badminton racquet and shuttlecock')
        first_match = mq.candidates('hedgehgo', spellcheck=True)[0]
        self.assertEqual(first_match.phrase, '🦔')
        self.assertEqual(first_match.comment, 'hedgehog')

    def test_candidates_various_unicode_chars_classic(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US', 'it_IT', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'],
            match_algorithm='classic')
        first_match, second_match = mq.candidates('euro sign')[:2]
        self.assertEqual(first_match.phrase, '💶')
        self.assertEqual(first_match.comment, 'banknote with euro sign')
        self.assertEqual(second_match.phrase, '€')
        self.assertEqual(second_match.comment, 'euro sign')
        first_match = mq.candidates('superscript one')[0]
        self.assertEqual(first_match.phrase, '¹')
        self.assertEqual(first_match.comment, 'superscript one')
        first_match, second_match = mq.candidates('currency sign')[:2]
        self.assertEqual(first_match.phrase, '¤')
        self.assertEqual(first_match.comment, 'currency sign {Currency}')
        self.assertEqual(second_match.phrase, '💴')
        self.assertEqual(second_match.comment, 'banknote with yen sign [currency]')
        first_match = mq.candidates('connector')[0]
        self.assertEqual(first_match.phrase, '﹎')
        self.assertEqual(first_match.comment, 'centreline low line {Connector}')
        first_match = mq.candidates('dash')[0]
        self.assertEqual(first_match.phrase, '💨')
        self.assertEqual(first_match.comment, 'dash symbol')
        first_match = mq.candidates('close')[0]
        self.assertEqual(first_match.phrase, '😥')
        self.assertEqual(first_match.comment, 'disappointed but relieved face [close]')
        first_match = mq.candidates('punctuation dotted')[0]
        self.assertEqual(first_match.phrase, '⁜')
        self.assertEqual(first_match.comment, 'dotted cross {Punctuation}')
        first_match = mq.candidates('final double quote')[0]
        self.assertEqual(first_match.phrase, '”')
        self.assertEqual(first_match.comment, 'right double quotation mark {Final quote}')
        first_match = mq.candidates('initial double quote')[0]
        self.assertEqual(first_match.phrase, '“')
        self.assertEqual(first_match.comment, 'left double quotation mark {Initial quote}')
        first_match = mq.candidates('modifier')[0]
        self.assertEqual(first_match.phrase, '🏻')
        self.assertEqual(first_match.comment, 'emoji modifier fitzpatrick type-1-2 {Modifier}')
        first_match = mq.candidates('math nabla')[0]
        self.assertEqual(first_match.phrase, '∇')
        self.assertEqual(first_match.comment, 'nabla {Math}')
        first_match = mq.candidates('separator line')[0]
        self.assertEqual(first_match.phrase, ' ')
        self.assertEqual(first_match.comment, 'U+2028 line separator {Line}')
        first_match = mq.candidates('separator paragraph')[0]
        self.assertEqual(first_match.phrase, ' ')
        self.assertEqual(first_match.comment, 'U+2029 paragraph separator {Paragraph}')
        first_match = mq.candidates('separator space')[0]
        self.assertEqual(first_match.phrase, ' ')
        self.assertEqual(first_match.comment, 'U+0020 space {Space}')

    @unittest.skipUnless(
        itb_emoji.IMPORT_RAPIDFUZZ_SUCCESSFUL,
        'Skipping because this test requires rapidfuzz to work.')
    def test_candidates_various_unicode_chars_rapidfuzz(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US', 'it_IT', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'],
            match_algorithm='rapidfuzz')
        first_match, second_match = mq.candidates('euro sign')[:2]
        self.assertEqual(first_match.phrase, '💶')
        self.assertEqual(first_match.comment, 'banknote with euro sign')
        self.assertEqual(second_match.phrase, '€')
        self.assertEqual(second_match.comment, 'euro sign')
        first_match = mq.candidates('superscript one')[0]
        self.assertEqual(first_match.phrase, '¹')
        self.assertEqual(first_match.comment, 'superscript one')
        first_match, second_match = mq.candidates('currency sign')[:2]
        self.assertEqual(first_match.phrase, '💲')
        self.assertEqual(first_match.comment, 'heavy dollar sign')
        self.assertEqual(second_match.phrase, '💴')
        self.assertEqual(second_match.comment, 'banknote with yen sign [currency]')
        first_match = mq.candidates('connector')[0]
        self.assertEqual(first_match.phrase, '_')
        self.assertEqual(first_match.comment, 'low line {Connector}')
        first_match = mq.candidates('dash')[0]
        self.assertEqual(first_match.phrase, '💨')
        self.assertEqual(first_match.comment, 'dash symbol [dashing]')
        first_match = mq.candidates('close')[0]
        self.assertEqual(first_match.phrase, '📕')
        self.assertEqual(first_match.comment, 'closed book')
        first_match = mq.candidates('punctuation')[0]
        self.assertEqual(first_match.phrase, '❗')
        self.assertEqual(first_match.comment, 'heavy exclamation mark symbol [punctuation]')
        first_match = mq.candidates('final double quote')[0]
        self.assertEqual(first_match.phrase, '”')
        self.assertEqual(first_match.comment, 'right double quotation mark {Final quote}')
        first_match = mq.candidates('initial double quote')[0]
        self.assertEqual(first_match.phrase, '“')
        self.assertEqual(first_match.comment, 'left double quotation mark {Initial quote}')
        first_match = mq.candidates('emoji modifier')[0]
        self.assertEqual(first_match.phrase, '🏻')
        self.assertEqual(first_match.comment, 'emoji modifier fitzpatrick type-1-2 {Modifier}')
        first_match = mq.candidates('math nabla')[0]
        self.assertEqual(first_match.phrase, '∇')
        self.assertEqual(first_match.comment, 'nabla {Math}')
        first_match = mq.candidates('separator line')[0]
        self.assertEqual(first_match.phrase, ' ')
        self.assertEqual(first_match.comment, 'U+2028 line separator {Line}')
        first_match = mq.candidates('separator paragraph')[0]
        self.assertEqual(first_match.phrase, ' ')
        self.assertEqual(first_match.comment, 'U+2029 paragraph separator {Paragraph}')
        first_match = mq.candidates('separator space')[0]
        self.assertEqual(first_match.phrase, ' ')
        self.assertEqual(first_match.comment, 'U+0020 space {Space}')

    def test_candidates_french_text_classic(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['fr_FR'], match_algorithm='classic')
        first_match = mq.candidates('chat')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'chat')
        first_match = mq.candidates('réflexion')[0]
        self.assertEqual(first_match.phrase, '🤔')
        self.assertEqual(first_match.comment, 'visage en pleine réflexion')

    @unittest.skipUnless(
        itb_emoji.IMPORT_RAPIDFUZZ_SUCCESSFUL,
        'Skipping because this test requires rapidfuzz to work.')
    def test_candidates_french_text_rapidfuzz(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['fr_FR'], match_algorithm='rapidfuzz')
        first_match = mq.candidates('chat animal')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'chat')
        first_match = mq.candidates('réflexion')[0]
        self.assertEqual(first_match.phrase, '🤔')
        self.assertEqual(first_match.comment, 'visage en pleine réflexion')

    def test_candidates_french_similar(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['fr_FR'])
        matches = mq.candidates('🤔', match_limit = 3)
        self.assertEqual(matches[0].phrase, '🤔')
        self.assertEqual(matches[0].comment, 'visage en pleine réflexion [🤔, émoticône, hum, méditer, penser, réfléchir, réflexion, visage, visage en pleine réflexion]')
        self.assertEqual(matches[0].user_freq, 9.0)
        self.assertEqual(matches[1].phrase, '🤐')
        self.assertEqual(matches[1].comment, 'visage avec bouche fermeture éclair [émoticône, visage]')
        self.assertEqual(matches[1].user_freq, 2.0)
        self.assertEqual(matches[2].phrase, '🤗')
        self.assertEqual(matches[2].comment, 'visage qui fait un câlin [émoticône, visage]')
        self.assertEqual(matches[2].user_freq, 2.0)

    def test_candidates_code_point_input_classic(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['fr_FR'], match_algorithm='classic')
        first_match = mq.candidates('2019')[0]
        self.assertEqual(first_match.phrase, '’')
        self.assertEqual(first_match.comment, 'U+2019 apostrophe droite')
        first_match = mq.candidates('41')[0]
        self.assertEqual(first_match.phrase, 'A')
        self.assertEqual(first_match.comment, 'U+41 latin capital letter a')
        first_match = mq.candidates('2a')[0]
        self.assertEqual(first_match.phrase, '*')
        self.assertEqual(first_match.comment, 'U+2A astérisque')
        first_match = mq.candidates('1b')[0]
        self.assertEqual(first_match.phrase, '\x1b')
        self.assertEqual(first_match.comment, 'U+1B')

    @unittest.skipUnless(
        itb_emoji.IMPORT_RAPIDFUZZ_SUCCESSFUL,
        'Skipping because this test requires rapidfuzz to work.')
    def test_candidates_code_point_input_rapidfuzz(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['fr_FR'], match_algorithm='rapidfuzz')
        first_match = mq.candidates('2019')[0]
        self.assertEqual(first_match.phrase, '’')
        self.assertEqual(first_match.comment, 'U+2019 apostrophe droite')
        first_match = mq.candidates('41')[0]
        self.assertEqual(first_match.phrase, 'A')
        self.assertEqual(first_match.comment, 'U+41 latin capital letter a')
        first_match = mq.candidates('2a')[0]
        self.assertEqual(first_match.phrase, '*')
        self.assertEqual(first_match.comment, 'U+2A astérisque')
        first_match = mq.candidates('1b')[0]
        self.assertEqual(first_match.phrase, '\x1b')
        self.assertEqual(first_match.comment, 'U+1B')

    def test_candidates_de_DE_versus_de_CH_classic(self) -> None: # pylint: disable=invalid-name
        mq = itb_emoji.EmojiMatcher(
            languages = ['de_DE'],
            match_algorithm = 'classic')
        first_match = mq.candidates('Reissverschluss')[0]
        self.assertEqual(first_match.phrase, '🤐')
        self.assertEqual(first_match.comment, 'Gesicht mit Reißverschlussmund')
        first_match = mq.candidates('Reißverschluss')[0]
        self.assertEqual(first_match.phrase, '🤐')
        self.assertEqual(first_match.comment, 'Gesicht mit Reißverschlussmund')
        first_match = mq.candidates('Reißverschluß')[0]
        self.assertEqual(first_match.phrase, '🤐')
        self.assertEqual(first_match.comment, 'Gesicht mit Reißverschlussmund')
        mq = itb_emoji.EmojiMatcher(
            languages = ['de_CH'],
            match_algorithm = 'classic')
        first_match = mq.candidates('Reissverschluss')[0]
        self.assertEqual(first_match.phrase, '🤐')
        self.assertEqual(first_match.comment, 'Gesicht mit Reissverschlussmund')
        first_match = mq.candidates('Reißverschluss')[0]
        self.assertEqual(first_match.phrase, '🤐')
        self.assertEqual(first_match.comment, 'Gesicht mit Reissverschlussmund')
        first_match = mq.candidates('Reißverschluß')[0]
        self.assertEqual(first_match.phrase, '🤐')
        self.assertEqual(first_match.comment, 'Gesicht mit Reissverschlussmund')

    @unittest.skipUnless(
        itb_emoji.IMPORT_RAPIDFUZZ_SUCCESSFUL,
        'Skipping because this test requires rapidfuzz to work.')
    def test_candidates_de_DE_versus_de_CH_rapidfuzz(self) -> None: # pylint: disable=invalid-name
        mq = itb_emoji.EmojiMatcher(
            languages = ['de_DE'],
            match_algorithm = 'rapidfuzz')
        first_match = mq.candidates('Reissverschluss')[0]
        self.assertEqual(first_match.phrase, '🤐')
        self.assertEqual(first_match.comment, 'Gesicht mit Reißverschlussmund')
        first_match = mq.candidates('Reißverschluss')[0]
        self.assertEqual(first_match.phrase, '🤐')
        self.assertEqual(first_match.comment, 'Gesicht mit Reißverschlussmund')
        first_match = mq.candidates('Reißverschluß')[0]
        self.assertEqual(first_match.phrase, '🤐')
        self.assertEqual(first_match.comment, 'Gesicht mit Reißverschlussmund')
        mq = itb_emoji.EmojiMatcher(
            languages = ['de_CH'],
            match_algorithm = 'rapidfuzz')
        first_match = mq.candidates('Reissverschluss')[0]
        self.assertEqual(first_match.phrase, '🤐')
        self.assertEqual(first_match.comment, 'Gesicht mit Reissverschlussmund')
        first_match = mq.candidates('Reißverschluss')[0]
        self.assertEqual(first_match.phrase, '🤐')
        self.assertEqual(first_match.comment, 'Gesicht mit Reissverschlussmund')
        first_match = mq.candidates('Reißverschluß')[0]
        self.assertEqual(first_match.phrase, '🤐')
        self.assertEqual(first_match.comment, 'Gesicht mit Reissverschlussmund')

    @unittest.skipIf(
        itb_emoji.IMPORT_PINYIN_SUCCESSFUL,
        "Skipping because import pinyin worked.")
    def test_candidates_pinyin_missing_zh_CN(self) -> None: # pylint: disable=invalid-name
        mq = itb_emoji.EmojiMatcher(
            languages = ['zh_CN'])
        first_match = mq.candidates('赛马')[0]
        self.assertEqual(first_match.phrase, '🏇')
        self.assertEqual(first_match.comment, '赛马')
        self.assertEqual(
            0, len(mq.candidates('saima')))

    @unittest.skipUnless(
        itb_emoji.IMPORT_PINYIN_SUCCESSFUL,
        "Skipping because import pinyin failed.")
    def test_candidates_pinyin_available_zh_CN_classic(self) -> None: # pylint: disable=invalid-name
        mq = itb_emoji.EmojiMatcher(
            languages = ['zh_CN'],
            match_algorithm = 'classic')
        first_match = mq.candidates('赛马')[0]
        self.assertEqual(first_match.phrase, '🏇')
        self.assertEqual(first_match.comment, '赛马')
        first_match = mq.candidates('saima')[0]
        self.assertEqual(first_match.phrase, '🏇')
        self.assertEqual(first_match.comment, '赛马 “sàimǎ”')

    @unittest.skipUnless(
        itb_emoji.IMPORT_PINYIN_SUCCESSFUL,
        "Skipping because import pinyin failed.")
    def test_candidates_pinyin_available_zh_CN_rapidfuzz(self) -> None: # pylint: disable=invalid-name
        mq = itb_emoji.EmojiMatcher(
            languages = ['zh_CN'],
            match_algorithm = 'rapidfuzz')
        first_match = mq.candidates('赛马')[0]
        self.assertEqual(first_match.phrase, '🏇')
        self.assertEqual(first_match.comment, '赛马')
        first_match = mq.candidates('saima')[0]
        self.assertEqual(first_match.phrase, '🏇')
        self.assertEqual(first_match.comment, '赛马 “sàimǎ” [qímǎ]')

    @unittest.skipIf(
        itb_emoji.IMPORT_PINYIN_SUCCESSFUL,
        "Skipping because import pinyin worked.")
    def test_candidates_pinyin_missing_zh_TW_classic(self) -> None: # pylint: disable=invalid-name
        mq = itb_emoji.EmojiMatcher(
            languages = ['zh_TW'],
            match_algorithm = 'classic')
        first_match = mq.candidates('賽馬')[0]
        self.assertEqual(first_match.phrase, '🏇')
        self.assertEqual(first_match.comment, '賽馬')
        self.assertEqual(
            0, len(mq.candidates('saima')))

    @unittest.skipIf(
        itb_emoji.IMPORT_PINYIN_SUCCESSFUL,
        "Skipping because import pinyin worked.")
    def test_candidates_pinyin_missing_zh_TW_rapidfuzz(self) -> None: # pylint: disable=invalid-name
        mq = itb_emoji.EmojiMatcher(
            languages = ['zh_TW'],
            match_algorithm = 'rapidfuzz')
        first_match = mq.candidates('賽馬')[0]
        self.assertEqual(first_match.phrase, '🏇')
        self.assertEqual(first_match.comment, '賽馬')
        self.assertEqual(
            0, len(mq.candidates('saima')))

    @unittest.skipUnless(
        itb_emoji.IMPORT_PINYIN_SUCCESSFUL,
        "Skipping because import pinyin failed.")
    def test_candidates_pinyin_available_zh_TW_classic(self) -> None: # pylint: disable=invalid-name
        mq = itb_emoji.EmojiMatcher(
            languages = ['zh_TW'],
            match_algorithm = 'classic')
        first_match = mq.candidates('賽馬')[0]
        self.assertEqual(first_match.phrase, '🏇')
        self.assertEqual(first_match.comment, '賽馬')
        first_match = mq.candidates('saima')[0]
        self.assertEqual(first_match.phrase, '🏇')
        self.assertEqual(first_match.comment, '賽馬 “sàimǎ”')

    @unittest.skipUnless(
        itb_emoji.IMPORT_PINYIN_SUCCESSFUL,
        "Skipping because import pinyin failed.")
    def test_candidates_pinyin_available_zh_TW_rapidfuzz(self) -> None: # pylint: disable=invalid-name
        mq = itb_emoji.EmojiMatcher(
            languages = ['zh_TW'],
            match_algorithm = 'rapidfuzz')
        first_match = mq.candidates('賽馬')[0]
        self.assertEqual(first_match.phrase, '🏇')
        self.assertEqual(first_match.comment, '賽馬')
        first_match = mq.candidates('saima')[0]
        self.assertEqual(first_match.phrase, '🏇')
        self.assertEqual(first_match.comment, '賽馬 “sàimǎ” [qímǎ]')

    @unittest.skipIf(
        itb_emoji.IMPORT_PYKAKASI_SUCCESSFUL,
        "Skipping because import pykakasi worked.")
    def test_candidates_pykakasi_missing_ja_JP_classic(self) -> None: # pylint: disable=invalid-name
        mq = itb_emoji.EmojiMatcher(
            languages = ['ja_JP'],
            match_algorithm = 'classic')
        self.assertEqual(
            0, len(mq.candidates('katatsumuri')))
        first_match = mq.candidates('かたつむり')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり')
        first_match = mq.candidates('かたつむり_')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり')
        first_match = mq.candidates('かたつむり＿')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり')
        first_match = mq.candidates('カタツムリ')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり [カタツムリ]')
        first_match = mq.candidates('カタツムリ_')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり [カタツムリ]')
        first_match = mq.candidates('カタツムリ＿')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり [カタツムリ]')
        first_match = mq.candidates('ネコ')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'ネコ')
        first_match = mq.candidates('ネコ_')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'ネコ')
        first_match = mq.candidates('ネコ＿')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'ネコ')
        self.assertEqual(
            0, len(mq.candidates('ねこ')))
        self.assertEqual(
            0, len(mq.candidates('ねこ_')))
        self.assertEqual(
            0, len(mq.candidates('ねこ＿')))

    @unittest.skipIf(
        itb_emoji.IMPORT_PYKAKASI_SUCCESSFUL,
        "Skipping because import pykakasi worked.")
    def test_candidates_pykakasi_missing_ja_JP_rapidfuzz(self) -> None: # pylint: disable=invalid-name
        mq = itb_emoji.EmojiMatcher(
            languages = ['ja_JP'],
            match_algorithm = 'rapidfuzz')
        self.assertEqual(
            0, len(mq.candidates('katatsumuri')))
        first_match = mq.candidates('かたつむり')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり')
        first_match = mq.candidates('かたつむり_')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり')
        first_match = mq.candidates('かたつむり＿')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり')
        first_match = mq.candidates('カタツムリ')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり [カタツムリ]')
        first_match = mq.candidates('カタツムリ_')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり [カタツムリ]')
        first_match = mq.candidates('カタツムリ＿')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり [カタツムリ]')
        first_match = mq.candidates('ネコ')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'ネコ')
        first_match = mq.candidates('ネコ_')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'ネコ')
        first_match = mq.candidates('ネコ＿')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'ネコ')
        self.assertEqual(
            0, len(mq.candidates('ねこ')))
        self.assertEqual(
            0, len(mq.candidates('ねこ_')))
        self.assertEqual(
            0, len(mq.candidates('ねこ＿')))

    @unittest.skipUnless(
        itb_emoji.IMPORT_PYKAKASI_SUCCESSFUL,
        "Skipping because import pykakasi failed.")
    def test_candidates_pykakasi_available_ja_JP_classic(self) -> None: # pylint: disable=invalid-name
        mq = itb_emoji.EmojiMatcher(
            languages = ['ja_JP'],
            match_algorithm = 'classic')
        first_match = mq.candidates('katatsumuri')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり “katatsumuri”')
        first_match = mq.candidates('かたつむり')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり')
        first_match = mq.candidates('かたつむり_')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり')
        first_match = mq.candidates('かたつむり＿')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり')
        first_match = mq.candidates('カタツムリ')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり [カタツムリ]')
        first_match = mq.candidates('カタツムリ_')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり [カタツムリ]')
        first_match = mq.candidates('カタツムリ＿')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり [カタツムリ]')
        first_match = mq.candidates('ネコ')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'ネコ')
        first_match = mq.candidates('ネコ_')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'ネコ')
        first_match = mq.candidates('ネコ＿')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'ネコ')
        first_match = mq.candidates('ねこ')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'ネコ “ねこ”')
        first_match = mq.candidates('ねこ_')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'ネコ “ねこ”')
        first_match = mq.candidates('ねこ＿')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'ネコ “ねこ”')

    @unittest.skipUnless(
        itb_emoji.IMPORT_PYKAKASI_SUCCESSFUL,
        "Skipping because import pykakasi failed.")
    def test_candidates_pykakasi_available_ja_JP_rapidfuzz(self) -> None: # pylint: disable=invalid-name
        mq = itb_emoji.EmojiMatcher(
            languages = ['ja_JP'],
            match_algorithm = 'rapidfuzz')
        first_match = mq.candidates('katatsumuri')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり “katatsumuri”')
        first_match = mq.candidates('かたつむり')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり')
        first_match = mq.candidates('かたつむり_')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり')
        first_match = mq.candidates('かたつむり＿')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり')
        first_match = mq.candidates('カタツムリ')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり [カタツムリ]')
        first_match = mq.candidates('カタツムリ_')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり [カタツムリ]')
        first_match = mq.candidates('カタツムリ＿')[0]
        self.assertEqual(first_match.phrase, '🐌')
        self.assertEqual(first_match.comment, 'かたつむり [カタツムリ]')
        first_match = mq.candidates('ネコ')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'ネコ')
        first_match = mq.candidates('ネコ_')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'ネコ')
        first_match = mq.candidates('ネコ＿')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'ネコ')
        first_match = mq.candidates('ねこ')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'ネコ “ねこ”')
        first_match = mq.candidates('ねこ_')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'ネコ “ねこ”')
        first_match = mq.candidates('ねこ＿')[0]
        self.assertEqual(first_match.phrase, '🐈')
        self.assertEqual(first_match.comment, 'ネコ “ねこ”')

if __name__ == '__main__':
    LOG_HANDLER = logging.StreamHandler(stream=sys.stderr)
    LOGGER.setLevel(logging.DEBUG)
    LOGGER.addHandler(LOG_HANDLER)
    unittest.main()



#!/usr/bin/python3

# ibus-typing-booster - A completion input method for IBus
#
# Copyright (c) 2019 Mike FABIAN <mfabian@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

'''
This file implements test cases for finding similar emojis
'''

import sys
import os
import logging
import unittest

LOGGER = logging.getLogger('ibus-typing-booster')

# pylint: disable=wrong-import-position
sys.path.insert(0, "../engine")
import itb_emoji # pylint: disable=import-error
sys.path.pop(0)
# pylint: enable=wrong-import-position

# Set the domain name to something invalid to avoid using
# the translations for the doctest tests. Translations may
# make the tests fail just because some translations are
# added, changed, or missing.
itb_emoji.DOMAINNAME = ''

# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=invalid-name
# pylint: disable=line-too-long

@unittest.skipIf(
    os.path.abspath(os.path.join(os.path.dirname(__file__), '../data/annotations'))
    != os.path.abspath(os.path.dirname(itb_emoji.find_cldr_annotation_path('en'))),
    f'Using external emoji annotations: '
    f'{itb_emoji.find_cldr_annotation_path("en")} '
    f'Testing with older emoji annotations instead '
    f'of those included in the ibus-typing-booster source is likely '
    f'to create meaningless test failures.')
class EmojiSimilarTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.maxDiff = None
        LOGGER.info("itb_emoji.find_cldr_annotation_path('en')->%s",
                    itb_emoji.find_cldr_annotation_path('en'))

    def tearDown(self) -> None:
        pass

    def test_dummy(self) -> None:
        self.assertEqual(True, True)

    def test_similar_query_is_not_an_emoji(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US', 'it_IT', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'])
        self.assertEqual(
            mq.similar('this is not an emoji', match_limit=5),
            [])

    def test_similar_white_smiling_face_en_US(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US', 'it_IT', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'])
        matches = mq.similar('☺', match_limit=5)
        self.assertEqual(matches[0].phrase, '☺️')
        self.assertEqual(matches[0].comment, 'white smiling face [☺️, So, people, face, outlined, relaxed, smile, uc1, happy, smiling]')
        self.assertEqual(matches[0].user_freq, 10.0)
        self.assertEqual(matches[1].phrase, '🥲')
        self.assertEqual(matches[1].comment,  'smiling face with tear [So, people, face, happy, smile, smiling]')
        self.assertEqual(matches[1].user_freq, 6.0)
        self.assertEqual(matches[2].phrase, '😇')
        self.assertEqual(matches[2].comment, 'smiling face with halo [So, people, face, smile, happy, smiling]')
        self.assertEqual(matches[2].user_freq, 6.0)
        self.assertEqual(matches[3].phrase, '🙂')
        self.assertEqual(matches[3].comment, 'slightly smiling face [So, people, face, smile, happy, smiling]')
        self.assertEqual(matches[3].user_freq, 6.0)
        self.assertEqual(matches[4].phrase, '😆')
        self.assertEqual(matches[4].comment, 'smiling face with open mouth and tightly-closed eyes [So, people, face, smile, happy, smiling]')
        self.assertEqual(matches[4].user_freq, 6.0)

    def test_similar_white_smiling_face_it_IT(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['it_IT', 'en_US', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'])
        matches = mq.similar('☺', match_limit=5)
        self.assertEqual(matches[0].phrase, '☺️')
        self.assertEqual(matches[0].comment, 'faccina sorridente [☺️, contorno faccina sorridente, delineata, emozionarsi, faccina, felice, rilassata, sorridente]')
        self.assertEqual(matches[0].user_freq, 8.0)
        self.assertEqual(matches[1].phrase, '😊')
        self.assertEqual(matches[1].comment, 'faccina con occhi sorridenti [faccina, felice]')
        self.assertEqual(matches[1].user_freq, 2.0)
        self.assertEqual(matches[2].phrase, '🙂')
        self.assertEqual(matches[2].comment, 'faccina con sorriso accennato [faccina, felice]')
        self.assertEqual(matches[2].user_freq, 2.0)
        self.assertEqual(matches[3].phrase, '😂')
        self.assertEqual(matches[3].comment, 'faccina con lacrime di gioia [faccina, felice]')
        self.assertEqual(matches[3].user_freq, 2.0)
        self.assertEqual(matches[4].phrase, '😃')
        self.assertEqual(matches[4].comment, 'faccina con sorriso e occhi spalancati [felice, sorridente]')
        self.assertEqual(matches[4].user_freq, 2.0)

    def test_similar_camel_en_US(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US', 'it_IT', 'es_MX', 'es_ES', 'de_DE', 'ja_JP'])
        matches = mq.similar('🐫', match_limit=5)
        self.assertEqual(matches[0].phrase, '🐫')
        self.assertEqual(matches[0].comment, 'bactrian camel [🐫, So, nature, bactrian, camel, hump, uc6, animal, desert, two, two-hump]')
        self.assertEqual(matches[0].user_freq, 11.0)
        self.assertEqual(matches[1].phrase, '🐪')
        self.assertEqual(matches[1].comment, 'dromedary camel [So, nature, hump, uc6, animal, camel, desert]')
        self.assertEqual(matches[1].user_freq, 7.0)
        self.assertEqual(matches[2].phrase, '🐌')
        self.assertEqual(matches[2].comment, 'snail [So, nature, uc6, animal, nature]')
        self.assertEqual(matches[2].user_freq, 5.0)
        self.assertEqual(matches[3].phrase, '🐝')
        self.assertEqual(matches[3].comment, 'honeybee [So, nature, uc6, animal, nature]')
        self.assertEqual(matches[3].user_freq, 5.0)
        self.assertEqual(matches[4].phrase,'🐞')
        self.assertEqual(matches[4].comment, 'lady beetle [So, nature, uc6, animal, nature]')
        self.assertEqual(matches[4].user_freq, 5.0)

    def test_similar_camel_it_IT(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['it_IT', 'en_US','es_MX', 'es_ES', 'de_DE', 'ja_JP'])
        matches = mq.similar('🐫', match_limit=5)
        self.assertEqual(matches[0].phrase, '🐫')
        self.assertEqual(matches[0].comment, 'cammello [🐫, animale, animali, cammello, deserto, due gobbe, gobba]')
        self.assertEqual(matches[0].user_freq, 7.0)
        self.assertEqual(matches[1].phrase, '🐪')
        self.assertEqual(matches[1].comment, 'dromedario [animale, animali, cammello, deserto, gobba]')
        self.assertEqual(matches[1].user_freq, 5.0)
        self.assertEqual(matches[2].phrase, '🐐')
        self.assertEqual(matches[2].comment, 'capra [animale, animali]')
        self.assertEqual(matches[2].user_freq, 2.0)
        self.assertEqual(matches[3].phrase, '🦒')
        self.assertEqual(matches[3].comment, 'giraffa [animale, animali]')
        self.assertEqual(matches[3].user_freq, 2.0)
        self.assertEqual(matches[4].phrase, '🐏')
        self.assertEqual(matches[4].comment, 'montone [animale, animali]')
        self.assertEqual(matches[4].user_freq, 2.0)

    def test_similar_camel_de_DE(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['de_DE', 'it_IT', 'en_US','es_MX', 'es_ES', 'ja_JP'])
        matches = mq.similar('🐫', match_limit=5)
        self.assertEqual(matches[0].phrase, '🐫')
        self.assertEqual(matches[0].comment, 'Kamel [🐫, Kamel, Tier, Wüste, zweihöckrig]')
        self.assertEqual(matches[0].user_freq, 5.0)
        self.assertEqual(matches[1].phrase, '🐪')
        self.assertEqual(matches[1].comment, 'Dromedar [Kamel, Tier, Wüste]')
        self.assertEqual(matches[1].user_freq, 3.0)
        self.assertEqual(matches[2].phrase, '🦙')
        self.assertEqual(matches[2].comment, 'Lama [Kamel, Tier]')
        self.assertEqual(matches[2].user_freq, 2.0)
        self.assertEqual(matches[3].phrase, '🐐')
        self.assertEqual(matches[3].comment, 'Ziege [Tier]')
        self.assertEqual(matches[3].user_freq, 1.0)
        self.assertEqual(matches[4].phrase, '🐑')
        self.assertEqual(matches[4].comment, 'Schaf [Tier]')
        self.assertEqual(matches[4].user_freq, 1.0)

    def test_similar_camel_es_MX(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['es_MX', 'it_IT', 'de_DE', 'en_US', 'es_ES', 'ja_JP'])
        matches = mq.similar('🐫', match_limit=5)
        self.assertEqual(matches[0].phrase, '🐫')
        self.assertEqual(matches[0].comment, 'camello [🐫, animal, camélido, camello, joroba]')
        self.assertEqual(matches[0].user_freq, 5.0)
        self.assertEqual(matches[1].phrase, '🐪')
        self.assertEqual(matches[1].comment, 'dromedario [animal, camélido, joroba]')
        self.assertEqual(matches[1].user_freq, 3.0)
        self.assertEqual(matches[2].phrase, '\U0001f999')
        self.assertEqual(matches[2].comment, 'llama [camélido]')
        self.assertEqual(matches[2].user_freq, 1.0)
        self.assertEqual(matches[3].phrase, '🐐')
        self.assertEqual(matches[3].comment, 'cabra [animal]')
        self.assertEqual(matches[3].user_freq, 1.0)
        self.assertEqual(matches[4].phrase, '🐑')
        self.assertEqual(matches[4].comment, 'oveja [animal]')
        self.assertEqual(matches[4].user_freq, 1.0)

    def test_similar_camel_es_ES(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['es_ES',  'it_IT', 'es_MX', 'de_DE', 'en_US', 'ja_JP'])
        matches = mq.similar('🐫', match_limit=5)
        self.assertEqual(matches[0].phrase, '🐫')
        self.assertEqual(matches[0].comment, 'camello [🐫, bactriano, camello, desierto, dromedario, jorobas]')
        self.assertEqual(matches[0].user_freq, 6.0)
        self.assertEqual(matches[1].phrase, '🐪')
        self.assertEqual(matches[1].comment, 'dromedario [camello, desierto, dromedario]')
        self.assertEqual(matches[1].user_freq, 3.0)
        self.assertEqual(matches[2].phrase, '🏜️')
        self.assertEqual(matches[2].comment, 'desierto [desierto]')
        self.assertEqual(matches[2].user_freq, 1.0)
        self.assertEqual(matches[3].phrase, '🐫')
        self.assertEqual(matches[3].comment, 'cammello [🐫, animale, animali, cammello, deserto, due gobbe, gobba]')
        self.assertEqual(matches[3].user_freq, 7.0)
        self.assertEqual(matches[4].phrase, '🐪')
        self.assertEqual(matches[4].comment, 'dromedario [animale, animali, cammello, deserto, gobba]')
        self.assertEqual(matches[4].user_freq, 5.0)

    def test_similar_euro_sign_es_ES(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['es_ES',  'it_IT', 'es_MX', 'de_DE', 'en_US', 'ja_JP'])
        matches = mq.similar('€', match_limit=5)
        self.assertEqual(matches[0].phrase, '€')
        self.assertEqual(matches[0].comment, 'euro [€, divisa, EUR, euro, moneda]')
        self.assertEqual(matches[0].user_freq, 5.0)
        self.assertEqual(matches[1].phrase, '£')
        self.assertEqual(matches[1].comment, 'libra esterlina [divisa, moneda]')
        self.assertEqual(matches[1].user_freq, 2.0)
        self.assertEqual(matches[2].phrase, '₽')
        self.assertEqual(matches[2].comment, 'rublo [divisa, moneda]')
        self.assertEqual(matches[2].user_freq, 2.0)
        self.assertEqual(matches[3].phrase, '₹')
        self.assertEqual(matches[3].comment, 'rupia india [divisa, moneda]')
        self.assertEqual(matches[3].user_freq, 2.0)
        self.assertEqual(matches[4].phrase, '¥')
        self.assertEqual(matches[4].comment, 'yen [divisa, moneda]')
        self.assertEqual(matches[4].user_freq, 2.0)

    def test_similar_surfer_es_ES(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['es_ES',  'it_IT', 'es_MX', 'de_DE', 'en_US', 'ja_JP'])
        matches = mq.similar('🏄‍♂', match_limit = 2)
        self.assertEqual(matches[0].phrase, '🏄\u200d♂️')
        self.assertEqual(matches[0].comment, 'hombre haciendo surf [🏄\u200d♂️, hombre, hombre haciendo surf, surf, surfero, surfista]')
        self.assertEqual(matches[0].user_freq, 6.0)
        self.assertEqual(matches[1].phrase, '🏄🏻\u200d♂️')
        self.assertEqual(matches[1].comment, 'hombre haciendo surf: tono de piel claro [hombre, hombre haciendo surf, surf, surfero, surfista]')
        self.assertEqual(matches[1].user_freq, 5.0)

    def test_similar_de_DE_versus_de_CH(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['de_DE'])
        matches = mq.similar('🤐', match_limit = 3)
        self.assertEqual(matches[0].phrase, '🤐')
        self.assertEqual(matches[0].comment, 'Gesicht mit Reißverschlussmund [🤐, Geheimnis, Gesicht, halten, Mund, Reißverschluss, schweigen, Smiley, wahren]')
        self.assertEqual(matches[0].user_freq, 9.0)
        self.assertEqual(matches[1].phrase, '😶')
        self.assertEqual(matches[1].comment, 'Gesicht ohne Mund [Gesicht, Mund, Smiley]')
        self.assertEqual(matches[1].user_freq, 3.0)
        self.assertEqual(matches[2].phrase, '🫢')
        self.assertEqual(matches[2].comment, 'Gesicht mit offenen Augen und Hand über dem Mund [Gesicht, Mund, Smiley]')
        self.assertEqual(matches[2].user_freq, 3.0)
        mq = itb_emoji.EmojiMatcher(
            languages = ['de_CH'])
        matches = mq.similar('🤐', match_limit = 3)
        self.assertEqual(matches[0].phrase, '🤐')
        self.assertEqual(matches[0].comment, 'Gesicht mit Reissverschlussmund [🤐, Geheimnis, Gesicht, halten, Mund, Reissverschluss, schweigen, Smiley, wahren]')
        self.assertEqual(matches[0].user_freq, 9.0)
        self.assertEqual(matches[1].phrase, '😅')
        self.assertEqual(matches[1].comment, 'grinsendes Gesicht mit Schweisstropfen [Gesicht, Mund, Smiley]')
        self.assertEqual(matches[1].user_freq, 3.0)
        self.assertEqual(matches[2].phrase, '😃')
        self.assertEqual(matches[2].comment, 'grinsendes Gesicht mit grossen Augen [Gesicht, Mund, Smiley]')
        self.assertEqual(matches[2].user_freq, 3.0)

    def test_similar_show_keywords_option_en_US(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['en_US'])
        matches =  mq.similar('🐌', match_limit = 3)
        self.assertEqual(matches[0].phrase, '🐌')
        self.assertEqual(matches[0].comment, 'snail [🐌, So, nature, snail, uc6, animal, escargot, garden, nature, slug]')
        self.assertEqual(matches[0].user_freq, 10.0)
        self.assertEqual(matches[1].phrase, '🐞')
        self.assertEqual(matches[1].comment, 'lady beetle [So, nature, uc6, animal, garden, nature]')
        self.assertEqual(matches[1].user_freq, 6.0)
        self.assertEqual(matches[2].phrase, '🐛')
        self.assertEqual(matches[2].comment, 'bug [So, nature, uc6, animal, garden]')
        self.assertEqual(matches[2].user_freq, 5.0)
        matches = mq.similar('🐌', match_limit = 3, show_keywords=False)
        self.assertEqual(matches[0].phrase, '🐌')
        self.assertEqual(matches[0].comment, 'snail')
        self.assertEqual(matches[0].user_freq, 10.0)
        self.assertEqual(matches[1].phrase, '🐞')
        self.assertEqual(matches[1].comment, 'lady beetle')
        self.assertEqual(matches[1].user_freq, 6.0)
        self.assertEqual(matches[2].phrase, '🐛')
        self.assertEqual(matches[2].comment, 'bug',)
        self.assertEqual(matches[2].user_freq, 5.0)

    @unittest.skipIf(
        itb_emoji.IMPORT_PINYIN_SUCCESSFUL,
        "Skipping because import pinyin worked.")
    def test_similar_horse_racing_pinyin_missing_zh_CN(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['zh_CN'])
        matches = mq.similar('🏇', match_limit = 3)
        self.assertEqual(matches[0].phrase, '🏇')
        self.assertEqual(matches[0].comment, '赛马 [🏇, 三冠, 赛马, 赛马骑师, 马, 骑师, 骑马]')
        self.assertEqual(matches[0].user_freq, 7.0)
        self.assertEqual(matches[1].phrase, '🏇🏻')
        self.assertEqual(matches[1].comment, '赛马: 较浅肤色 [三冠, 赛马, 赛马骑师, 马, 骑师, 骑马]')
        self.assertEqual(matches[1].user_freq, 6.0)
        self.assertEqual(matches[2].phrase, '🏇🏼')
        self.assertEqual(matches[2].comment, '赛马: 中等-浅肤色 [三冠, 赛马, 赛马骑师, 马, 骑师, 骑马]')
        self.assertEqual(matches[2].user_freq, 6.0)

    @unittest.skipUnless(
        itb_emoji.IMPORT_PINYIN_SUCCESSFUL,
        "Skipping because import pinyin failed.")
    def test_similar_horse_racing_pinyin_available_zh_CN(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['zh_CN'])
        matches = mq.similar('🏇', match_limit = 3)
        self.assertEqual(matches[0].phrase, '🏇')
        self.assertEqual(matches[0].comment, '赛马 [🏇, 三冠, sānguān, 赛马, sàimǎ, 赛马骑师, sàimǎqíshī, 马, mǎ, 骑师, qíshī, 骑马, qímǎ]')
        self.assertEqual(matches[0].user_freq, 13.0)
        self.assertEqual(matches[1].phrase, '🏇🏻')
        self.assertEqual(matches[1].comment, '赛马: 较浅肤色 [三冠, sānguān, 赛马, sàimǎ, 赛马骑师, sàimǎqíshī, 马, mǎ, 骑师, qíshī, 骑马, qímǎ]')
        self.assertEqual(matches[1].user_freq, 12.0)
        self.assertEqual(matches[2].phrase, '🏇🏼')
        self.assertEqual(matches[2].comment, '赛马: 中等-浅肤色 [三冠, sānguān, 赛马, sàimǎ, 赛马骑师, sàimǎqíshī, 马, mǎ, 骑师, qíshī, 骑马, qímǎ]')
        self.assertEqual(matches[2].user_freq, 12.0)

    @unittest.skipIf(
        itb_emoji.IMPORT_PINYIN_SUCCESSFUL,
        "Skipping because import pinyin worked.")
    def test_similar_horse_racing_pinyin_missing_zh_TW(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['zh_TW'])
        matches = mq.similar('🏇', match_limit = 1)
        self.assertEqual(matches[0].phrase, '🏇')
        self.assertEqual(matches[0].comment, '賽馬 [🏇, 賽馬, 馬, 騎馬]')
        self.assertEqual(matches[0].user_freq, 4.0)

    @unittest.skipUnless(
        itb_emoji.IMPORT_PINYIN_SUCCESSFUL,
        "Skipping because import pinyin failed.")
    def test_similar_horse_racing_pinyin_available_zh_TW(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['zh_TW'])
        matches = mq.similar('🏇', match_limit = 1)
        self.assertEqual(matches[0].phrase, '🏇')
        self.assertEqual(matches[0].comment, '賽馬 [🏇, 賽馬, sàimǎ, 馬, mǎ, 騎馬, qímǎ]')
        self.assertEqual(matches[0].user_freq, 7.0)

    @unittest.skipIf(
        itb_emoji.IMPORT_PYKAKASI_SUCCESSFUL,
        "Skipping because import pykakasi worked.")
    def test_candidates_pykakasi_missing_ja_JP(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['ja_JP'])
        matches = mq.similar('🐤', match_limit=5)
        self.assertEqual(matches[0].phrase, '🐤')
        self.assertEqual(matches[0].comment, 'ひよこ [🐤, ひな, ひよこ, 動物, 横を向いているひよこ, 顔, 鳥]')
        self.assertEqual(matches[0].user_freq, 7.0)
        self.assertEqual(matches[1].phrase, '🐣',)
        self.assertEqual(matches[1].comment, '卵からかえったひよこ [ひな, ひよこ, 動物, 顔, 鳥]')
        self.assertEqual(matches[1].user_freq, 5.0)
        self.assertEqual(matches[2].phrase, '🐥')
        self.assertEqual(matches[2].comment, '前を向いているひよこ [ひな, ひよこ, 動物, 鳥]')
        self.assertEqual(matches[2].user_freq, 4.0)
        self.assertEqual(matches[3].phrase, '🐦')
        self.assertEqual(matches[3].comment, '鳥 [動物, 顔, 鳥]')
        self.assertEqual(matches[3].user_freq, 3.0)
        self.assertEqual(matches[4].phrase, '🐔')
        self.assertEqual(matches[4].comment, 'にわとり [動物, 顔, 鳥]')
        self.assertEqual(matches[4].user_freq, 3.0)
        matches = mq.similar('🐌', match_limit=5)
        self.assertEqual(matches[0].phrase, '🐌')
        self.assertEqual(matches[0].comment, 'かたつむり [🐌, エスカルゴ, かたつむり, カタツムリ, でんでん虫, 虫]')
        self.assertEqual(matches[0].user_freq, 6.0)
        self.assertEqual(matches[1].phrase, '🦋')
        self.assertEqual(matches[1].comment, 'チョウ [虫]')
        self.assertEqual(matches[1].user_freq, 1.0)
        self.assertEqual(matches[2].phrase, '🐛')
        self.assertEqual(matches[2].comment, '毛虫 [虫]')
        self.assertEqual(matches[2].user_freq, 1.0)
        self.assertEqual(matches[3].phrase, '🐜')
        self.assertEqual(matches[3].comment, 'アリ [虫]')
        self.assertEqual(matches[3].user_freq, 1.0)
        self.assertEqual(matches[4].phrase, '🐝')
        self.assertEqual(matches[4].comment, 'ミツバチ [虫]')
        self.assertEqual(matches[4].user_freq, 1.0)
        matches = mq.similar('😱', match_limit=5)
        self.assertEqual(matches[0].phrase, '😱')
        self.assertEqual(matches[0].comment, '恐怖 [😱, がーん, ショック, 叫び, 叫んでいる顔, 恐怖, 顔, 驚き]')
        self.assertEqual(matches[0].user_freq, 8.0)
        self.assertEqual(matches[1].phrase, '🤯')
        self.assertEqual(matches[1].comment, '頭爆発 [ショック, 顔, 驚き]')
        self.assertEqual(matches[1].user_freq, 3.0)
        self.assertEqual(matches[2].phrase, '🙀')
        self.assertEqual(matches[2].comment, '絶望する猫 [がーん, ショック, 顔]')
        self.assertEqual(matches[2].user_freq, 3.0)
        self.assertEqual(matches[3].phrase, '😨')
        self.assertEqual(matches[3].comment, '青ざめ [がーん, 顔]')
        self.assertEqual(matches[3].user_freq, 2.0)
        self.assertEqual(matches[4].phrase, '😧')
        self.assertEqual(matches[4].comment, '苦悩 [顔, 驚き]')
        self.assertEqual(matches[4].user_freq, 2.0)

    @unittest.skipUnless(
        itb_emoji.IMPORT_PYKAKASI_SUCCESSFUL,
        "Skipping because import pykakasi failed.")
    def test_candidates_pykakasi_available_ja_JP(self) -> None:
        mq = itb_emoji.EmojiMatcher(
            languages = ['ja_JP'])
        matches = mq.similar('🐤', match_limit=5)
        self.assertEqual(matches[0].phrase, '🐤')
        self.assertEqual(matches[0].comment, 'ひよこ [🐤, ひな, ひよこ, 動物, どうぶつ, 横を向いているひよこ, よこをむいているひよこ, 顔, かお, 鳥, とり, hina, hiyoko, doubutsu, yokowomuiteiruhiyoko, kao, tori]')
        self.assertEqual(matches[0].user_freq, 17.0)
        self.assertEqual(matches[1].phrase, '🐣')
        self.assertEqual(matches[1].comment, '卵からかえったひよこ [ひな, ひよこ, 動物, どうぶつ, 顔, かお, 鳥, とり, hina, hiyoko, doubutsu, kao, tori]')
        self.assertEqual(matches[1].user_freq, 13.0)
        self.assertEqual(matches[2].phrase, '🐥')
        self.assertEqual(matches[2].comment, '前を向いているひよこ [ひな, ひよこ, 動物, どうぶつ, 鳥, とり, hina, hiyoko, doubutsu, tori]')
        self.assertEqual(matches[2].user_freq, 10.0)
        self.assertEqual(matches[3].phrase, '🐦')
        self.assertEqual(matches[3].comment, '鳥 [動物, どうぶつ, 顔, かお, 鳥, とり, doubutsu, kao, tori]')
        self.assertEqual(matches[3].user_freq, 9.0)
        self.assertEqual(matches[4].phrase, '🐔')
        self.assertEqual(matches[4].comment, 'にわとり [動物, どうぶつ, 顔, かお, 鳥, とり, doubutsu, kao, tori]')
        self.assertEqual(matches[4].user_freq, 9.0)
        matches = mq.similar('🐌', match_limit=5)
        self.assertEqual(matches[0].phrase, '🐌')
        self.assertEqual(matches[0].comment, 'かたつむり [🐌, エスカルゴ, えすかるご, かたつむり, カタツムリ, でんでん虫, でんでんむし, 虫, むし, esukarugo, katatsumuri, dendenmushi, mushi]')
        self.assertEqual(matches[0].user_freq, 13.0)
        self.assertEqual(matches[1].phrase, '🦋')
        self.assertEqual(matches[1].comment, 'チョウ [虫, むし, mushi]')
        self.assertEqual(matches[1].user_freq, 3.0)
        self.assertEqual(matches[2].phrase, '🐛')
        self.assertEqual(matches[2].comment, '毛虫 [虫, むし, mushi]')
        self.assertEqual(matches[2].user_freq, 3.0)
        self.assertEqual(matches[3].phrase, '🐜')
        self.assertEqual(matches[3].comment, 'アリ [虫, むし, mushi]')
        self.assertEqual(matches[3].user_freq, 3.0)
        self.assertEqual(matches[4].phrase, '🐝')
        self.assertEqual(matches[4].comment, 'ミツバチ [虫, むし, mushi]')
        self.assertEqual(matches[4].user_freq, 3.0)
        matches = mq.similar('😱', match_limit=5)
        self.assertEqual(matches[0].phrase, '😱')
        self.assertEqual(matches[0].comment, '恐怖 [😱, がーん, ショック, しょっく, 叫び, さけび, 叫んでいる顔, さけんでいるかお, 恐怖, きょうふ, 顔, かお, 驚き, おどろき, gaan, shokku, sakebi, sakendeirukao, kyoufu, kao, odoroki]')
        self.assertEqual(matches[0].user_freq, 21.0)
        self.assertEqual(matches[1].phrase, '🤯')
        self.assertEqual(matches[1].comment, '頭爆発 [ショック, しょっく, 顔, かお, 驚き, おどろき, shokku, kao, odoroki]')
        self.assertEqual(matches[1].user_freq, 9.0)
        self.assertEqual(matches[2].phrase, '🙀')
        self.assertEqual(matches[2].comment, '絶望する猫 [がーん, ショック, しょっく, 顔, かお, gaan, shokku, kao]')
        self.assertEqual(matches[2].user_freq, 8.0)
        self.assertEqual(matches[3].phrase, '🫨')
        self.assertEqual(matches[3].comment, '震えている顔 [がーん, ショック, しょっく, 顔, かお, gaan, shokku, kao]')
        self.assertEqual(matches[3].user_freq, 8.0)
        self.assertEqual(matches[4].phrase, '😧')
        self.assertEqual(matches[4].comment, '苦悩 [顔, かお, 驚き, おどろき, kao, odoroki]')
        self.assertEqual(matches[4].user_freq, 6.0)

if __name__ == '__main__':
    LOG_HANDLER = logging.StreamHandler(stream=sys.stderr)
    LOGGER.setLevel(logging.DEBUG)
    LOGGER.addHandler(LOG_HANDLER)
    unittest.main()



# ibus-typing-booster - A completion input method for IBus
#
# Copyright (c) 2016-2019 Mike FABIAN <mfabian@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
'''
 Define some mock classes for the unittests.
'''

from typing import Any
from typing import List
from typing import Optional
from typing import Callable
# pylint: disable=wrong-import-position
from gi import require_version
require_version('IBus', '1.0')
from gi.repository import IBus
# pylint: enable=wrong-import-position

def mock_glib_idle_add(
        func: Callable[..., Any],
        *args: Any,
        **kwargs: Any) -> Any:
    '''When unit testing, call the function immediately'''
    func(*args, **kwargs)
    return 0  # GLib.idle_add returns a source ID

def mock_glib_timeout_add(
        _interval: int,
        func: Callable[..., Any],
        *args: Any,
        **kwargs: Any) -> Any:
    '''When unit testing, call the function immediately'''
    func(*args, **kwargs)
    return 0  # GLib.timeout_add returns a source ID

# pylint: disable=missing-function-docstring
# pylint: disable=too-few-public-methods
# pylint: disable=missing-class-docstring
class MockEngine:
    '''Mock engine class to simulate IBus.Engine for testing'''
    props = None
    def __init__(
            self,
            engine_name: str = '',
            connection: Any = None,
            object_path: str = '') -> None:
        self.mock_engine_name = engine_name
        self.mock_connection = connection
        self.mock_object_path = object_path
        self.mock_auxiliary_text = ''
        self.mock_preedit_text = ''
        self.mock_preedit_text_cursor_pos = 0
        self.mock_preedit_text_visible = True
        self.mock_preedit_focus_mode = IBus.PreeditFocusMode.COMMIT
        self.mock_committed_text = ''
        self.mock_committed_text_cursor_pos = 0
        self.mock_committed_text_codes: List[int] = []
        self.mock_committed_text_states: List[int] = []
        self.client_capabilities = (
            IBus.Capabilite.PREEDIT_TEXT
            | IBus.Capabilite.AUXILIARY_TEXT
            | IBus.Capabilite.LOOKUP_TABLE
            | IBus.Capabilite.FOCUS
            | IBus.Capabilite.PROPERTY)
        # There are lots of weird problems with surrounding text
        # which makes this hard to test. Therefore this mock
        # engine does not try to support surrounding text, i.e.
        # we omit “| IBus.Capabilite.SURROUNDING_TEXT” here.

    def update_auxiliary_text(self, text: IBus.Text, _visible: bool) -> None:
        self.mock_auxiliary_text = text.text

    def commit_text(self, text: IBus.Text) -> None:
        self.mock_committed_text = (
            self.mock_committed_text[
                :self.mock_committed_text_cursor_pos]
            + text.text
            + self.mock_committed_text[
                self.mock_committed_text_cursor_pos:])
        self.mock_committed_text_codes = (
            self.mock_committed_text_codes[
                :self.mock_committed_text_cursor_pos]
            + [-1 for char in text.text] # -1 indicates commit, not forward
            + self.mock_committed_text_codes[
                self.mock_committed_text_cursor_pos:])
        self.mock_committed_text_states = (
            self.mock_committed_text_states[
                :self.mock_committed_text_cursor_pos]
            + [-1 for char in text.text] # -1 indicates commit, not forward
            + self.mock_committed_text_states[
                self.mock_committed_text_cursor_pos:])
        self.mock_committed_text_cursor_pos += len(text.text)

    def forward_key_event(self, val: int, code: int, state: int) -> None:
        if (val == IBus.KEY_Left
            and self.mock_committed_text_cursor_pos > 0):
            self.mock_committed_text_cursor_pos -= 1
            return
        unicode = IBus.keyval_to_unicode(val)
        if unicode:
            self.mock_committed_text = (
            self.mock_committed_text[
                :self.mock_committed_text_cursor_pos]
            + unicode
            + self.mock_committed_text[
                self.mock_committed_text_cursor_pos:])
            self.mock_committed_text_codes = (
                self.mock_committed_text_codes[
                    :self.mock_committed_text_cursor_pos]
                + [code]
                + self.mock_committed_text_codes[
                    self.mock_committed_text_cursor_pos:])
            self.mock_committed_text_states = (
                self.mock_committed_text_states[
                    :self.mock_committed_text_cursor_pos]
                + [state]
                + self.mock_committed_text_states[
                    self.mock_committed_text_cursor_pos:])
            self.mock_committed_text_cursor_pos += len(unicode)

    def update_lookup_table(
            self, table: IBus.LookupTable, visible: bool) -> None:
        pass

    def update_preedit_text(self, text: IBus.Text, cursor_pos: int, visible: bool) -> None:
        self.mock_preedit_text = text.get_text()
        self.mock_preedit_text_cursor_pos = cursor_pos
        self.mock_preedit_text_visible = visible

    def update_preedit_text_with_mode(
            self,
            text: IBus.Text,
            cursor_pos: int,
            visible: bool,
            focus_mode: IBus.PreeditFocusMode) -> None:
        self.mock_preedit_focus_mode = focus_mode
        self.update_preedit_text(text, cursor_pos, visible)

    def hide_preedit_text(self) -> None:
        self.update_preedit_text(IBus.Text.new_from_string(''), 0, False)

    def show_preedit_text(self) -> None:
        pass

    def register_properties(self, property_list: List[IBus.Property]) -> None:
        pass

    def update_property(self, _prop: IBus.Property) -> None:
        pass

    def hide_lookup_table(self) -> None:
        pass

    def get_surrounding_text(self) -> None:
        pass

    def connect(self, signal: str, callback_function: Any) -> None:
        pass

    def add_table_by_locale(self, locale: str) -> None:
        pass

class MockLookupTable:
    def __init__(
            self,
            page_size: int = 9,
            cursor_pos: int = 0,
            cursor_visible: bool = False,
            wrap_around: bool = True) -> None:
        self.clear()
        self.mock_labels = {
            1: '1.',
            2: '2.',
            3: '3.',
            4: '4.',
            5: '5.',
            6: '6.',
            7: '7.',
            8: '8.',
            9: '9.',
        }
        self.mock_page_size = page_size
        self.mock_cursor_pos = cursor_pos
        self.mock_cursor_visible = cursor_visible
        self.mock_wrap_around = wrap_around
        self.mock_candidates: List[str] = []
        self.mock_orientation = 0

    def clear(self) -> None:
        self.mock_candidates = []
        self.mock_cursor_pos = 0

    def set_label(self, index: int, label: IBus.Text) -> None:
        self.mock_labels[index] = label.get_text()

    def get_label(self, index: int) -> str:
        return self.mock_labels[index]

    def set_page_size(self, size: int) -> None:
        self.mock_page_size = size

    def get_page_size(self) -> int:
        return self.mock_page_size

    def set_round(self, wrap_around: bool) -> None:
        self.mock_wrap_around = wrap_around

    def set_cursor_pos(self, pos: int) -> None:
        self.mock_cursor_pos = pos

    def get_cursor_pos(self) -> int:
        return self.mock_cursor_pos

    def set_cursor_visible(self, visible: bool) -> None:
        self.mock_cursor_visible = visible

    def is_cursor_visible(self) -> bool:
        return self.mock_cursor_visible

    def cursor_down(self) -> None:
        if self.mock_candidates:
            self.mock_cursor_pos += 1
            self.mock_cursor_pos %= len(self.mock_candidates)

    def cursor_up(self) -> None:
        if self.mock_candidates:
            if self.mock_cursor_pos > 0:
                self.mock_cursor_pos -= 1
            else:
                self.mock_cursor_pos = len(self.mock_candidates) - 1

    def set_orientation(self, orientation: int) -> None:
        self.mock_orientation = orientation

    def get_orientation(self) -> int:
        return self.mock_orientation

    def get_number_of_candidates(self) -> int:
        return len(self.mock_candidates)

    def append_candidate(self, candidate: IBus.Text) -> None:
        self.mock_candidates.append(candidate.get_text())

    def get_candidate(self, index: int) -> str:
        return self.mock_candidates[index]

class MockPropList:
    def __init__(self) -> None:
        self.mock_prop_list: List[IBus.Property] = []

    def append(self, prop: IBus.Property) -> None:
        self.mock_prop_list.append(prop)

    def get(self, index: int) -> Optional[IBus.Property]:
        if len(self.mock_prop_list) < index + 1:
            return None
        return self.mock_prop_list[index]

class MockProperty:
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    def set_label(self, ibus_text: IBus.Text) -> None:
        pass

    def set_symbol(self, ibus_text: IBus.Text) -> None:
        pass

    def set_tooltip(self, ibus_text: IBus.Text) -> None:
        pass

    def set_sensitive(self, sensitive: bool) -> None:
        pass

    def set_visible(self, visible: bool) -> None:
        pass

    def set_state(self, visible: bool) -> None:
        pass



#!/usr/bin/python3

# ibus-typing-booster - A completion input method for IBus
#
# Copyright (c) 2019 Mike FABIAN <mfabian@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

'''
This file implements test cases for itb_pango.py
'''

from typing import TYPE_CHECKING
import sys
import os
import logging
import unittest

LOGGER = logging.getLogger('ibus-typing-booster')

IMPORT_DISTRO_SUCCESSFUL = False
try:
    import distro
    IMPORT_DISTRO_SUCCESSFUL = True
except (ImportError,):
    IMPORT_DISTRO_SUCCESSFUL = False

# pylint: disable=wrong-import-position,import-error
sys.path.insert(0, "../engine")
from itb_gtk import Gdk # type: ignore
if TYPE_CHECKING:
    # These imports are only for type checkers (mypy). They must not be
    # executed at runtime because itb_gtk controls the Gtk/Gdk versions.
    # pylint: disable=reimported
    from gi.repository import Gdk  # type: ignore
    # pylint: enable=reimported
import itb_pango
sys.path.pop(0)
# pylint: enable=wrong-import-position,import-error

# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

@unittest.skipUnless(
    'XDG_SESSION_TYPE' in os.environ
    and os.environ['XDG_SESSION_TYPE'] in ('x11', 'wayland'),
    'XDG_SESSION_TYPE is neither "x11" nor "wayland".')
@unittest.skipIf(Gdk.Display.open('') is None, 'Display cannot be opened.')
class ItbPangoTestCase(unittest.TestCase):
    def setUp(self) -> None:
        fonts_used = itb_pango.get_fonts_used_for_text('emoji', '😇', fallback=True)
        _run, results_for_run = fonts_used[0]
        self._fallback_font_name = results_for_run['font']
        LOGGER.info('Fallback font name=“%s”', self._fallback_font_name)
        # 🫩︎ U+1FAE9 FACE WITH BAGS UNDER EYES, added in Unicode 16.0
        fonts_used = itb_pango.get_fonts_used_for_text('emoji', '🫩', fallback=True)
        _run, results_for_run = fonts_used[0]
        self._fallback_font_name_u16 = results_for_run['font']
        LOGGER.info('Fallback font name Unicode 16.0=“%s”', self._fallback_font_name_u16)

    def tearDown(self) -> None:
        pass

    def font_available_or_skip(self, font_family_name: str) -> None:
        if font_family_name not in itb_pango.get_available_font_names():
            self.skipTest(f'{font_family_name} is not available.')

    def test_dummy(self) -> None:
        self.assertEqual(True, True)

    @unittest.skipUnless(
        IMPORT_DISTRO_SUCCESSFUL
        and distro.id() == 'fedora'
        and distro.version() >= '40',
        'Skipping, fonts might be different on Fedora < 40 or other distributions.')
    def test_dejavu_sans(self) -> None:
        font_name = 'DejaVu Sans'
        self.font_available_or_skip(font_name)
        self.font_available_or_skip(self._fallback_font_name)

        text = '' # empty string
        fallback = False
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 0)
        fallback = True
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 0)

        text = '\n' # new line
        fallback = False
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 0)
        fallback = True
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 0)

        text = '\u0008' # BACKSPACE
        fallback = False
        # supported, fallback will not be used:
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, True)
        self.assertEqual(results_for_run['glyph-available'], False)

        text = '\u001B' # ESCAPE
        fallback = False
        # supported, fallback will not be used:
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, True)
        self.assertEqual(results_for_run['glyph-available'], False)

        text = ' ' # normal SPACE
        fallback = False
        # supported, fallback will not be used:
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], False) # SPACE is not visible
        self.assertEqual('glyph-available' in results_for_run, True)
        self.assertEqual(results_for_run['glyph-available'], True)

        text = 'a' # normal letter
        fallback = False
        # supported, fallback will not be used:
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], True) # visible
        self.assertEqual('glyph-available' in results_for_run, True)
        self.assertEqual(results_for_run['glyph-available'], True)

    @unittest.skipUnless(
        IMPORT_DISTRO_SUCCESSFUL
        and distro.id() == 'fedora'
        and distro.version() >= '40',
        'Skipping, fonts might be different on Fedora < 40 or other distributions.')
    def test_twitter_color_emoji(self) -> None:
        font_name = 'Twitter Color Emoji'
        self.font_available_or_skip(font_name)
        self.font_available_or_skip(self._fallback_font_name)

        text = '☺\uFE0F' # request emoji representation
        fallback = False
        # supported, fallback will not be used:
        self.assertEqual(False, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 2)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, False)

        text = '☺\uFE0F' # request emoji representation
        fallback = True
        # supported, but with fallback enabled, Pango falls back nevertheless:
        self.assertEqual(False, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], self._fallback_font_name)

        text = '☺' # unqualified
        fallback = False
        # supported, fallback will not be used:
        self.assertEqual(False, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, True)
        self.assertEqual(results_for_run['glyph-available'], True)

        text = '☺' # unqualified
        fallback = True
        # supported,  and without the variation selector, Pango does not fall back:
        self.assertEqual(False, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, True)
        self.assertEqual(results_for_run['glyph-available'], True)

    @unittest.skipUnless(
        IMPORT_DISTRO_SUCCESSFUL
        and distro.id() == 'fedora'
        and distro.version() >= '40',
        'Skipping, fonts might be different on Fedora < 40 or other distributions.')
    def test_symbola(self) -> None:
        font_name = 'Symbola'
        self.font_available_or_skip(font_name)
        self.font_available_or_skip(self._fallback_font_name)

        text = '☺' # Lacks Emoji_Presentation, defaults to text representation
        fallback = False
        # supported by Symbola
        self.assertEqual(False, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, True)
        self.assertEqual(results_for_run['glyph-available'], True)

        text = '☺' # Lacks Emoji_Presentation, defaults to text representation
        fallback = True
        # supported by Symbola
        self.assertEqual(False, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, True)
        self.assertEqual(results_for_run['glyph-available'], True)

        text = '☺\uFE0E' # request text representation
        fallback = False
        # supported by Symbola
        self.assertEqual(False, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 2)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, False)

        text = '☺\uFE0E' # request text representation
        fallback = True
        # supported by Symbola, as Text representation is requested,
        # fallback to “Noto Color Emoji” does not happen:
        self.assertEqual(False, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 2)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, False)

        text = '☺\uFE0F' # request emoji representation
        fallback = False
        # supported by Symbola
        self.assertEqual(False, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 2)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, False)

        text = '☺\uFE0F' # request emoji representation
        fallback = True
        # supported by Symbola but request for emoji representation triggers fallback:
        self.assertEqual(False, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], self._fallback_font_name)

        text =  '😇' # unqualified, has Emoji_Presentation property
        fallback = False
        # Symbola is used:
        self.assertEqual(False, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, True)
        self.assertEqual(results_for_run['glyph-available'], True)

        text = '😇\uFE0E' # has Emoji_Presentation property
        fallback = False
        # Symbola is used:
        self.assertEqual(False, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 2)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, False)

        text = '😇' # unqualified, has Emoji_Presentation property
        fallback = True
        self.assertEqual(False, itb_pango.emoji_font_fallback_needed(font_name, text))
        # Symbola is **not** used:
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], self._fallback_font_name)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, True)
        self.assertEqual(results_for_run['glyph-available'], True)

        # 🫩︎ U+1FAE9 FACE WITH BAGS UNDER EYES, added in Unicode 16.0,
        # has Emoji_Presentation property
        text = '🫩' # unqualified
        fallback = False
        # Symbola has no glyph for this:
        self.assertEqual(True, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, True)
        self.assertEqual(results_for_run['glyph-available'], False)

        # 🫩︎ U+1FAE9 FACE WITH BAGS UNDER EYES, added in Unicode 16.0,
        # has Emoji_Presentation property
        text = '🫩\uFE0E'
        fallback = False
        # Symbola has no glyph for this:
        self.assertEqual(True, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 2)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, False)

        # 🫩︎ U+1FAE9 FACE WITH BAGS UNDER EYES, added in Unicode 16.0,
        # has Emoji_Presentation property
        text = '🫩' # unqualified
        fallback = True
        # fallback is used:
        self.assertEqual(True, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], self._fallback_font_name_u16)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, True)
        self.assertEqual(results_for_run['glyph-available'], True)

        # 🫩︎ U+1FAE9 FACE WITH BAGS UNDER EYES, added in Unicode 16.0,
        # has Emoji_Presentation property
        text = '🫩\uFE0E'
        fallback = True
        # fallback to “Noto Color Emoji” is used:
        self.assertEqual(True, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], self._fallback_font_name_u16)
        self.assertEqual(results_for_run['glyph-count'], 2)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, False)

    @unittest.skipUnless(
        IMPORT_DISTRO_SUCCESSFUL
        and distro.id() == 'fedora'
        and distro.version() >= '40',
        'Skipping, fonts might be different on Fedora < 40 or other distributions.')
    def test_twemoji(self) -> None:
        font_name = 'Twemoji'
        self.font_available_or_skip(font_name)
        self.font_available_or_skip(self._fallback_font_name)

        text = '☺\uFE0F'
        fallback = False
        # supported by Twemoji
        self.assertEqual(False, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, False)

        text = '🙂‍↕️'
        fallback = False
        # twitter-twemoji-fonts-14.0.2-5.fc40.noarch
        # does not support this and does not render this as a single glyph:
        self.assertEqual(True, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 3)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, False)
        fallback = True
        # Falling back to “Noto Color Emoji”
        self.assertEqual(True, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], self._fallback_font_name)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, False)

        text = '🇨🇶'
        fallback = False
        # “Twemoji” has no glyph for the flag of Sark (added in
        # Unicode 16.0) but “Noto Color Emoji” has it.  Even though
        # “Twemoji” has no glyph for the flag of Sark, Pango renders
        # the sequence of two code points (U+1F1E8 U+1F1F) as one
        # glyph when “Twemoji” is specified and fallback is not
        # allowed (Visually the glyph shown appears empty, there is no
        # “Tofu”):
        self.assertEqual(True, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], False) # it’s empty!
        self.assertEqual('glyph-available' in results_for_run, False)
        fallback = True
        # falling back:
        # If “OpenMoji Color” is the system default, it falls back to that
        # even though the current “OpenMoji Color” does not yet support Unicode 16.0
        # and renders the flag of Sark as two glyphs 🇨 U+1F1E8 🇶 U+1F1F6
        self.assertEqual(True, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], self._fallback_font_name)

        text = '🏴󠁧󠁢󠁷󠁬󠁳󠁿'
        fallback = False
        # “Twemoji” does correctly support the flag of Wales:
        self.assertEqual(False, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, False)
        fallback = True
        # with fallback enabled, Pango falls back to “Noto Color Emoji” nevertheless:
        self.assertEqual(False, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], self._fallback_font_name)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, False)

        # 🫩︎ U+1FAE9 FACE WITH BAGS UNDER EYES, added in Unicode 16.0
        text = '🫩'
        fallback = False
        # “Twemoji” does not have the glyph for this single code point
        # emoji, (visually the glyph shown when Twemoji is used is a
        # “Tofu” block with the code point inside):
        self.assertEqual(True, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, True)
        self.assertEqual(results_for_run['glyph-available'], False) # Tofu glyph shown
        fallback = True
        # with fallback enabled, Pango correctly falls back:
        self.assertEqual(True, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], self._fallback_font_name_u16)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, True)
        self.assertEqual(results_for_run['glyph-available'], True) # real glyph shown

        text = '🤥' # U+1F925, single code point
        fallback = False
        # “Twemoji” does correctly support it:
        self.assertEqual(False, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], font_name)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, True)
        self.assertEqual(results_for_run['glyph-available'], True) # real glyph shown
        fallback = True
        # with fallback enabled, Pango falls back to “Noto Color Emoji” nevertheless:
        self.assertEqual(False, itb_pango.emoji_font_fallback_needed(font_name, text))
        fonts_used = itb_pango.get_fonts_used_for_text(font_name, text, fallback=fallback)
        self.assertEqual(len(fonts_used), 1)
        run, results_for_run = fonts_used[0]
        self.assertEqual(run, text)
        self.assertEqual(results_for_run['font'], self._fallback_font_name)
        self.assertEqual(results_for_run['glyph-count'], 1)
        self.assertEqual(results_for_run['visible'], True)
        self.assertEqual('glyph-available' in results_for_run, True)
        self.assertEqual(results_for_run['glyph-available'], True) # real glyph shown

    @unittest.skipUnless(
        IMPORT_DISTRO_SUCCESSFUL
        and distro.id() == 'fedora'
        and distro.version() >= '40',
        'Skipping, fonts might be different on Fedora < 40 or other distributions.')
    def test_twemoji_fallback_needed(self) -> None:
        font_name = 'Twemoji'
        self.font_available_or_skip(font_name)
        # Twemoji does not support the emoji sequence for “head
        # shaking vertically” (U+1F642 U+200D U+2195, added in Unicode
        # 15.1):
        self.assertEqual(itb_pango.emoji_font_fallback_needed(font_name, '🙂‍↕️'), True)
        # Twemoji does not have the flag of Sark (U+1F1E8 U+1F1F6, added in Unicode 16.0):
        self.assertEqual(itb_pango.emoji_font_fallback_needed(font_name, '🇨🇶'), True)
        # Twemoji does not have U+1FAE9 FACE WITH BAGS UNDER EYES (added in Unicode 16.0):
        self.assertEqual(itb_pango.emoji_font_fallback_needed(font_name, '🫩'), True)
        # But Twemoji has U+1F925 LYING FACE (added in Unicode 9.0):
        self.assertEqual(itb_pango.emoji_font_fallback_needed(font_name, '🤥'), False)

    @unittest.skipUnless(
        IMPORT_DISTRO_SUCCESSFUL
        and distro.id() == 'fedora'
        and distro.version() >= '40',
        'Skipping, fonts might be different on Fedora < 40 or other distributions.')
    def test_twitter_color_emoji_fallback_needed(self) -> None:
        font_name = 'Twitter Color Emoji'
        self.font_available_or_skip(font_name)
        self.assertEqual(itb_pango.emoji_font_fallback_needed(font_name, '🙂‍↕️'), False)
        self.assertEqual(itb_pango.emoji_font_fallback_needed(font_name, '🤥'), False)
        self.assertEqual(itb_pango.emoji_font_fallback_needed(font_name, '☺\ufe0f'), False)
        self.assertEqual(itb_pango.emoji_font_fallback_needed(font_name, '☺\ufe0e'), False)
        # No regular characters:
        self.assertEqual(itb_pango.emoji_font_fallback_needed(font_name, 'A'), True)
        # Flag of Sark, Unicode 16.0, not available in 'Twitter Color Emoji' yet:
        self.assertEqual(itb_pango.emoji_font_fallback_needed(font_name, '🇨🇶'), True)
        # Fallback might always be needed for mor than one emoji, we don’t know:
        self.assertEqual(itb_pango.emoji_font_fallback_needed(font_name, '🏴󠁧󠁢󠁷󠁬󠁳󠁿🤥'), True)

if __name__ == '__main__':
    LOG_HANDLER = logging.StreamHandler(stream=sys.stderr)
    LOGGER.setLevel(logging.DEBUG)
    LOGGER.addHandler(LOG_HANDLER)
    unittest.main()
